Package: scanMiR
Type: Package
Title: scanMiR
Version: 1.19.0
Depends: R (>= 4.0)
Date: 2025-07-24
Authors@R: c(
    person("Pierre-Luc", "Germain", email="pierre-luc.germain@hest.ethz.ch", 
            role=c("cre","aut"), comment=c(ORCID="0000-0003-3418-4218")),
    person("Michael", "Soutschek", email="michael.soutschek@hest.ethz.ch", role="aut"),
    person("Fridolin", "Gross", email="fridolin.gross@u-bordeaux.fr", role="aut"))
Imports: Biostrings, pwalign, GenomicRanges, IRanges, data.table,
        BiocParallel, methods, Seqinfo, S4Vectors, ggplot2, stats,
        stringi, utils, graphics, grid, seqLogo, cowplot
Suggests: knitr, rmarkdown, BiocStyle, testthat (>= 3.0.0)
Description: 
    A set of tools for working with miRNA affinity models (KdModels),
    efficiently scanning for miRNA binding sites, and predicting target 
    repression. It supports scanning using miRNA seeds, full miRNA sequences 
    (enabling 3' alignment) and KdModels, and includes the prediction of slicing
    and TDMD sites. Finally, it includes utility and plotting functions (e.g. 
    for the visual representation of miRNA-target alignment).
License: GPL-3
VignetteBuilder: knitr
RoxygenNote: 7.3.1
biocViews: miRNA, SequenceMatching, Alignment
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/scanMiR
git_branch: devel
git_last_commit: a482f71
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-29
NeedsCompilation: no
Packaged: 2026-04-30 00:08:09 UTC; biocbuild
Author: Pierre-Luc Germain [cre, aut] (ORCID:
    <https://orcid.org/0000-0003-3418-4218>),
  Michael Soutschek [aut],
  Fridolin Gross [aut]
Maintainer: Pierre-Luc Germain <pierre-luc.germain@hest.ethz.ch>

# Generated by roxygen2: do not edit by hand

export(KdModelList)
export(aggregateMatches)
export(assignKdType)
export(conservation)
export(dummyKdData)
export(findSeedMatches)
export(get3pAlignment)
export(get8merRange)
export(getKdModel)
export(getKmers)
export(getMatchTypes)
export(getRandomSeq)
export(getSeed8mers)
export(plotKdModel)
export(removeOverlappingRanges)
export(viewTargetAlignment)
exportClasses(KdModel)
exportClasses(KdModelList)
exportMethods("[")
exportMethods(c)
exportMethods(show)
exportMethods(summary)
import(GenomicRanges)
import(ggplot2)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpnworkers)
importFrom(Biostrings,DNAString)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,RNAString)
importFrom(Biostrings,complement)
importFrom(Biostrings,consensusMatrix)
importFrom(Biostrings,extractAt)
importFrom(Biostrings,padAndClip)
importFrom(Biostrings,reverseComplement)
importFrom(Biostrings,subseq)
importFrom(Biostrings,xscat)
importFrom(Seqinfo,seqlevels)
importFrom(IRanges,FactorList)
importFrom(IRanges,IRanges)
importFrom(IRanges,RleList)
importFrom(S4Vectors,"mcols<-")
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,Rle)
importFrom(S4Vectors,head)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(cowplot,plot_grid)
importFrom(data.table,":=")
importFrom(data.table,.N)
importFrom(data.table,.SD)
importFrom(data.table,as.data.table)
importFrom(data.table,dcast)
importFrom(data.table,rbindlist)
importFrom(graphics,par)
importFrom(graphics,text)
importFrom(methods,is)
importFrom(methods,new)
importFrom(pwalign,nucleotideSubstitutionMatrix)
importFrom(pwalign,pairwiseAlignment)
importFrom(pwalign,pattern)
importFrom(pwalign,subject)
importFrom(pwalign,type)
importFrom(seqLogo,seqLogo)
importFrom(stats,.lm.fit)
importFrom(stats,cor)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stringi,stri_reverse)
importFrom(utils,data)
importFrom(utils,read.delim)

[![](https://img.shields.io/badge/doi-10.1093/bioinformatics/btac110-blue.svg)](https://doi.org/10.1093/bioinformatics/btac110) 
[![R build status](https://github.com/ETHZ-INS/scanMiR/workflows/R-CMD-check/badge.svg)](https://github.com/ETHZ-INS/scanMiR/actions)
<img align="right" style="margin-left: 20px; margin-bottom: 10px;" src="inst/docs/sticker.svg"/>

# scanMiR

`scanMiR` is a bioconductor set of tools for working with miRNA affinity
models (KdModels), enabling efficient and flexible scanning for miRNA binding 
sites and prediction of target repression.

This repository details the R package; for the web interface [click here](http://ethz-ins.org/scanMiR/).

<br/>

## Getting started

### Installation

```{r}
BiocManager::install("ETHZ-INS/scanMiR")
```

### Basic example usage

```{r}
# accepts miRNA target seeds:
findSeedMatches(seqs, seeds="AAACCAC")
# full miRNA sequences:
findSeedMatches(seqs, seeds="UUAAUGCUAAUCGUGAUAGGGGUU")
# or KdModels:
findSeedMatches(seqs, seeds=model)
```

```
GRanges object with 43 ranges and 4 metadata columns:
       seqnames    ranges strand |  p3.score          type    log_kd     note
          <Rle> <IRanges>  <Rle> | <integer>      <factor> <integer>    <Rle>
   [1]     seq2 2687-2694      * |         4 7mer-a1           -3607        -
   [2]     seq2 2358-2365      * |         0 6mer              -2341        -
   [3]     seq2 2550-2557      * |         0 6mer-m8            -986        -
   [4]     seq2 1642-1649      * |         0 non-canonical      -952        -
   [5]     seq2   920-927      * |         0 non-canonical      -847        -
   ...      ...       ...    ... .       ...           ...       ...      ...
```

The putative 3' binding of each match can also be visualized:

```{r}
viewTargetAlignment(matches[1], miRNA=model, seqs=seqs)
```
```
miRNA            3'-UUGGGGAUAGUGCUA---AUCGUAAUU-5'     
                              |||||     ||||||-        
target 5'-...NUAUAGACGAGUGACCUACGAUAUGCCGCAUUAAUU...-3'
```

<br/><br/><br/>

`scanMiR` can predict TDMD and circRNA slicing sites, and aggregate sites to 
predict repression based on the biochemical model from 
[McGeary, Lin et al. (2019)](https://dx.doi.org/10.1126/science.aav1741). For 
more information, see our [paper](https://doi.org/10.1093/bioinformatics/btac110).

To learn more about the functionalities, see the package's 
[vignettes](https://www.bioconductor.org/packages/devel/bioc/vignettes/scanMiR/inst/doc/scanning.html).
To obtain predicted `KdModels` for all mouse, human and rat miRbase miRNAs, see the 
[scanMiRData](https://github.com/ETHZ-INS/scanMiRData) pacakge.

For convenient wrappers connecting to AnnotationHub, fast out-of memory access 
to large collections, or a web interface to scanMiR, see the 
[scanMiRApp](https://github.com/ETHZ-INS/scanMiRApp) package.

#' getKmers
#'
#' Returns all combinations of `n` elements of `from`
#'
#' @param n Number of elements
#' @param from Letters sampled
#'
#' @return A character vector
#' @export
#'
#' @examples
#' getKmers(3)
getKmers <- function(n=4, from=c("A", "C", "G", "T")){
  apply(expand.grid(lapply(seq_len(n), FUN=function(x) from)),
        1, collapse="",FUN=paste)
}

#' getRandomSeq
#'
#' Produces a random sequence of the given letters
#'
#' @param length Length of the sequence
#' @param alphabet Letters from which to sample
#' @param n The number of sequences to generate
#'
#' @return A character vector of length 1
#' @export
#'
#' @examples
#' getRandomSeq(100)
getRandomSeq <- function(length=3000, alphabet=c("A","C","G","T"), n=1){
  seqs <- vapply(seq_len(n), FUN.VALUE=character(1), FUN=function(i){
    paste(sample(alphabet, size = length, replace=TRUE), collapse="")
  })
  names(seqs) <- paste0("seq",seq_len(n))
  seqs
}


#' getSeed8mers
#'
#' Generates all possible 8mers with 4 consecutive and positioned matches to a
#' given seed.
#'
#' @param seed The miRNA seed (target DNA sequence), a character vector of
#' length 8 (if of length 7, a "A" will be added on the right)
#' @param addNs Logical; whether to include 8mers with one flanking N
#'
#' @return A vector of 1024 8mers.
#' @export
#'
#' @examples
#' head(getSeed8mers("ACACTCCA"))
getSeed8mers <- function(seed, addNs=FALSE){
  a <- strsplit(as.character(seed),"")[[1]]
  if(length(a)==7) a <- c(a, "A")
  nts <- c("A","C","G","T")
  kmers <- expand.grid(nts,nts,nts,nts)
  y <- unique(unlist(lapply(0:4, FUN=function(x){
    paste0(
      do.call(paste0,kmers[,seq_len(x),drop=FALSE]),
      paste(a[x+1:4],collapse=""),
      do.call(paste0,kmers[,x+seq_len(4-x),drop=FALSE])
    )
  })))
  if(addNs){
    kmers <- expand.grid(nts,nts,nts)
    y <- c(y,unique(unlist(lapply(1:4, FUN=function(x){
      paste0("N",
             do.call(paste0,kmers[,seq_len(x-1),drop=FALSE]),
             paste(a[x+1:4],collapse=""),
             do.call(paste0,kmers[,x+seq_len(4-x)-1,drop=FALSE])
      )
    }))))
    y <- c(y,unique(unlist(lapply(0:3, FUN=function(x){
      paste0(do.call(paste0,kmers[,seq_len(x),drop=FALSE]),
             paste(a[x+1:4],collapse=""),
             do.call(paste0,kmers[,x+seq_len(3-x),drop=FALSE]),
             "N"
      )
    }))))
  }
  y
}

.build4mersRegEx <- function(seed){
  if(is(seed,"KdModel")){
    a <- strsplit(as.character(seed$canonical.seed),"")[[1]]
  }else{
    a <- strsplit(as.character(seed),"")[[1]]
  }
  if(length(a)==7) a <- c(a, "A")
  pats <- vapply(0:4, FUN.VALUE=character(1), FUN=function(x){
    paste0(paste(rep("[^N]",x), collapse=""),
           paste(a[x+1:4],collapse=""),
           paste(rep(".",4-x), collapse=""))
  })
  paste(pats,collapse="|")
}


.guessSeqType <- function(x, use.subset=TRUE){
  if(use.subset) seqs <- x[sample.int(length(x),min(length(x),10))]
  u <- any(grepl("U",seqs))
  t <- any(grepl("T",seqs))
  if(t && u) stop("Sequences contain both T and U!")
  if(t || u) return(ifelse(u,"RNA","DNA"))
  if(length(seqs)>10) return(.guessSeqType(x, FALSE))
  warning("Sequences contain neither T or U - assuming they are DNA...")
  return("DNA")
}

#' Create dummy log_kd per 12-mer data
#'
#' @param mod Optional model from which to create the dummy data
#'
#' @return A data.frame with 12-mers and log_kds
#' @export
#' @importFrom utils data
#' @examples
#' kd <- dummyKdData()
dummyKdData <- function(mod=NULL){
  if(is.null(mod)){
    data("SampleKdModel")
    mod <- SampleKdModel
  }
  mer12 <- paste0(getKmers(2),getSeed8mers(mod$canonical.seed),getKmers(2))
  data.frame(X12mer=mer12, log_kd=mod$mer8/1000)
}


#' get8merRange
#'
#' Returns the minimum and maximum 8-mer log-kd values
#'
#' @param mod A `KdModel`
#'
#' @return A numeric vector of length two
#' @export
#' @examples
#' data("SampleKdModel")
#' get8merRange(SampleKdModel)
get8merRange <- function(mod){
  stopifnot(is(mod,"KdModel"))
  mer8 <- getSeed8mers(mod$canonical.seed)
  mer8 <- which(mer8==mod$canonical.seed)
  fl <- rowSums(apply(.flankingValues(), 2, FUN=range))
  fl*mod$fl[mer8]+mod$mer8[mer8]
}

data("SampleKdModel", package="scanMiR")

kmers <- data.frame(
  type=c("8mer", "6mer-m8", "7mer-m8", "7mer-a1", "6mer", "6mer-a1",
           "g-bulged 7mer", "wobbled 8mer", "g-bulged 6mer", "wobbled 7mer"),
  log_kd=c(-5083L, -1550L, -4108L, -3749L, -2779L, -1624L, 97L, -1844L, 186L,
           -1216L),
  kmer = c("NNAGCATTAANN", "NNAGCATTCANN", "NNAGCATTACNN", "NNCGCATTAANN",
           "NNCGCATTACNN", "NNAACATTAANN", "NNGCGATTAANN", "NNAGTATTAANN",
           "NNGCGATTACNN", "NNAGTATTACNN")
)

test_that("Kd assignment works", {
  expect_type(assignKdType(kmers$kmer, SampleKdModel)$log_kd, "integer")
  expect_equal(assignKdType(kmers$kmer, SampleKdModel)$log_kd,
               kmers$log_kd, tolerance=200)
  expect_equal(get8merRange(SampleKdModel), c(-5831,-3978), tolerance=200)
})

test_that("Type assignment works", {
  expect_equal(as.character(assignKdType(kmers$kmer, SampleKdModel)$type),
               kmers$type)
})

mod <- getKdModel(kd=dummyKdData(), mirseq="TTAATGCTAATCGTGATAGGGGTT",
                  name="my-miRNA")

test_that("KdModel construction works", {
  expect_s4_class(mod, "KdModel")
  expect(all(!is.na(mod$mer8) & !is.infinite(mod$mer8)),
         "Missing/infinite values in `mer8`")
  expect(all(!is.na(mod$fl) & !is.infinite(mod$fl)),
         "Missing/infinite values in `fl`")
  expect_type(summary(mod), "character")
  expect_type(capture_output(show(mod)), "character")
  expect_gt(mod$cor,0.9)
  expect_gt(cor(mod$mer8, SampleKdModel$mer8),0.9)
  expect_identical(as.character(assignKdType(kmers$kmer, mod)$type),
               kmers$type)
  expect_identical(assignKdType(c("CTAGCATTAAGT","CTAGCATTACGT"), mod)$log_kd,
               c(-5083,-4108), tolerance=250)
})

test_that("KdModel plotting works",{
  expect_s3_class(plotKdModel(SampleKdModel, what="seeds"), "ggplot")
})

test_that("KdModelList operations work", {
  expect_s4_class(KdModelList(mod,SampleKdModel), "KdModelList")
  ml <- c(mod,SampleKdModel)
  expect_s4_class(ml, "KdModelList")
  expect_s4_class(ml[[1]], "KdModel")
  expect_s4_class(ml$`my-miRNA`, "KdModel")
  expect_s4_class(ml[2:1], "KdModelList")
  expect_equal(names(ml), c("my-miRNA", "hsa-miR-155-5p"))
})

#' @exportClass KdModelList
#' @rdname KdModelList
setClass(
  "KdModelList",
  contains="list",
  validity=function(object){
    if(!all(vapply(object, is, logical(1), class2="KdModel")))
      stop("A KdModelList should be a list of objects of class 'KdModel'")
    if(any(duplicated(names(object))))
      stop("There are multiple KdModels with the same name!")
    mn <- as.character(unlist(lapply(object, FUN=function(x) x$name)))
    if(!identical(names(object),mn)) stop("Mismatched names!")
    return(TRUE)
  }
)

#' KdModelList
#'
#' @param ... Any number of \code{\link{KdModel}} objects or lists thereof.
#' @param description A description for the collection.
#' @param makeUnique Logical; whether to rename models if names are duplicated.
#'
#' @return A KdModelList
#' @export
#' @examples
#' data(SampleKdModel)
#' mods <- KdModelList(SampleKdModel, SampleKdModel, makeUnique = TRUE)
#' mods
KdModelList <- function(..., description=NULL, makeUnique=FALSE){
  x <- list(...)
  isKdm <- vapply(x, is, logical(1), class2="KdModel")
  if(any(!isKdm)){
    warn <- FALSE
    for(f in which(!isKdm)){
      y <- x[[f]]
      x <- x[-f]
      if(is(y,"KdModelList") ||
         (is.list(y) && all(vapply(y, is, logical(1), class2="KdModel")))){
        x <- c(x,as.list(y))
      }else{
        warn <- TRUE
      }
    }
    if(warn) warning("Some objects were not KdModels and were discarded.")
  }
  nn <- vapply(x, FUN.VALUE=character(1), FUN=function(x) x$name)
  if(length(wdup <- which(duplicated(nn)))>0){
    if(makeUnique){
      wdup <- which(nn %in% unique(nn[wdup]))
      nn <- make.unique(nn)
      for(i in wdup) x[[i]]$name <- nn[i]
    }else{
      stop("There are multiple KdModels with the same name. ",
           "Use `KdModelList(..., makeUnique=TRUE)` to automatically rename ",
           "them.")
    }
  }
  names(x) <- nn
  x <- new("KdModelList", x)
  if(!is.null(attr(x, "created"))) attr(x, "created") <- Sys.Date()
  if(!is.null(description)) attr(x, "description") <- description
  x
}

#' Methods for the \code{\link{KdModelList}} classes
#' @name KdModelList-methods
#' @rdname KdModelList-methods
#' @aliases KdModelList-methods,KdModelList-method
#' @seealso \code{\link{KdModel}}, \code{\link{KdModelList}}
#' @param object,x An object of class \code{\link{KdModelList}}
#' @return Depends on the method.
#' @examples
#' # create a KdModelList :
#' data(SampleKdModel)
#' kml <- KdModelList( SampleKdModel, SampleKdModel, makeUnique=TRUE )
#' summary(kml)
#' kml[1] # returns a KdModelList
#' kml[[2]] # returns a KdModel
#' conservation(kml)
NULL

#' @rdname KdModelList-methods
#' @aliases KdModelList-methods,KdModelList-method
#' @export
setMethod("summary", "KdModelList", function(object){
  d <- attr(object, "created")
  cat(paste0("A `KdModelList` object",
            ifelse(is.null(d), "", paste0(" created on ",d,",\n")),
            " containing binding affinity models from ", length(object),
            " miRNAs.\n"))
  if(!is.null(desc <- attr(object, "description"))) cat(paste0(desc,"\n"))
  cons <- conservation(object)
  if(!all(is.na(cons))){
    print(table(conservation(object), useNA="ifany"))
  }
})

#' @rdname KdModelList-methods
#' @aliases [,KdModelList-methods,KdModelList-method
#' @export
#' @param i the index of item(s) to select
#' @param j,drop,... ignored
setMethod("[", "KdModelList", function(x, i, j=NULL, ..., drop = TRUE){
  xo <- new("KdModelList", unclass(x)[i])
  if(!is.null(attr(x, "created"))) attr(xo, "created") <- attr(x, "created")
  if(!is.null(attr(x, "description")))
    attr(xo, "description") <- attr(x, "description")
  xo
})


#' conservation
#'
#' @param x A KdModelList, or a KdModel
#'
#' @return A vector of the conservation status for each miRNA
#' @export
#' @examples
#' data(SampleKdModel)
#' conservation(SampleKdModel)
conservation <- function(x){
  lvls <- .conservation_levels()
  if(is(x,"KdModelList")){
    y <- factor(vapply(x, FUN.VALUE=integer(1), FUN=function(x){
      if(is.null(x$conservation)) return(NA_integer_)
      x$conservation
    }), levels=names(lvls))
  }else if(is(x,"KdModel")){
    if(is.null(x$conservation)) return(NA)
    y <- factor(x$conservation, levels=names(lvls))
  }else{
    stop("Undefined for an object of class ", class(x))
  }
  levels(y) <- as.character(lvls)
  y
}

.conservation_levels <- function(){
  c("-1"="Low-confidence", "0"="Poorly conserved",
    "1"="Conserved across mammals", "2"="Conserved across vertebrates")
}

#' plotKdModel
#'
#' Plots the summary of an affinity model.
#'
#' @param mod A `KdModel`
#' @param what Either 'seeds', 'logo', or 'both' (default).
#' @param n The number of top 7-mers to plot (when `what='seeds'`)
#'
#' @return If `what="logo"`, returns nothing and plots a position weight matrix.
#' Otherwise returns a ggplot.
#'
#' @details
#' `what='seeds'` plots the -$log(K_d)$ values of the top `n` 7-mers (including
#' both canonical and non-canonical sites), with or without the final "A"
#' vis-a-vis the first miRNA nucleotide.
#' `what='logo'` plots a `seqLogo` (requires the
#' [seqLogo]{https://bioconductor.org/packages/release/bioc/html/seqLogo.html}
#' package) showing the nucleotide-wise information content and preferences for
#' all 12-mers (centered around the seed, oriented in the direction of the target
#' mRNA). `what="both"` plots both.
#' Note that if the package `ggseqlogo` is installed, this will be used instead
#' to plot the logo, resulting in more detailed plot annotation.
#'
#' @import ggplot2
#' @importFrom cowplot plot_grid
#' @importFrom seqLogo seqLogo
#' @export
#' @examples
#' data(SampleKdModel)
#' plotKdModel(SampleKdModel, what="seeds")
plotKdModel <- function(mod, what=c("both","seeds","logo"), n=10){
  stopifnot(is(mod,"KdModel"))
  what <- match.arg(what)
  if(what=="seeds"){
    type_cols <- .typeColors()
    mirseq2 <- strsplit(gsub("T","U",mod$mirseq),"")[[1]]
    mirseq2 <- paste0("3'-",stringi::stri_reverse(gsub("T","U",mod$mirseq)),"-5'")
    mer8 <- getSeed8mers(mod$canonical.seed)
    wA <- which(substr(mer8,8,8)=="A")
    mer7 <- substr(mer8,1,7)
    As <- mod$mer8[wA]
    names(As) <- mer7[wA]
    mer.mean <- rowsum(mod$mer8[-wA],mer7[-wA])[,1]/3
    As <- As-mer.mean[names(As)]
    d <- data.frame(seed=names(mer.mean), base=mer.mean/-1000,
                    "A"=As[names(mer.mean)]/-1000,
                    type=getMatchTypes(names(mer.mean),mod$canonical.seed),
                    row.names=NULL)
    d <- d[head(order(d$base+d$A, decreasing=TRUE),n=n),]
    d$seed <- gsub("T","U",d$seed)
    d$seed <- factor(as.character(d$seed), rev(as.character(d$seed)))
    levels(d$type) <- .matchLevels(FALSE)
    d2 <- data.frame(seed=rep(d$seed,2), log_kd=c(d$base,d$A),
                     type=c(as.character(d$type), rep("+A",n)))
    p <- ggplot(d2, aes(seed, log_kd, fill=type)) + geom_col() + coord_flip() +
      ylab(bquote("-"*log(K[D]))) + xlab("7-mer") + ggtitle(mod$name) + 
      theme_minimal() +
      theme(axis.text.x = element_text(size=11),
            axis.text.y = element_text(size=11, family="mono", face = "bold"),
            axis.title.x = element_text(size=14),
            axis.title.y = element_text(size=14))
    p <- p + scale_fill_manual(values = type_cols[p$data$type])
    if(mod$name != mod$mirseq) p <- p + labs(subtitle=mirseq2)
    return( p )
  }

  if(what=="logo"){
#    if(requireNamespace("ggseqlogo", quietly=TRUE)){
#      p <- suppressWarnings(ggseqlogo( mod$pwm )) + 
#        labs(y="Information (bits)", x="miRNA 5' position")
#      p$scales$scales[[1]] <- scale_x_continuous(breaks=1:12,labels=c(10:1,"",""))
#      p <- p + 
#      theme(axis.text.x=element_text(size=11), axis.text.y=element_text(size=11),
#            axis.title.x=element_text(size=14), axis.title.y=element_text(size=14))
#    }else{
      p <- plot_grid(grid::grid.grabExpr(seqLogo( mod$pwm )))
#    }
    return(p)
  }
  plot_grid(plotKdModel(mod, "seeds"),plotKdModel(mod, "logo"),
            ncol=1, rel_heights = c(6,4))
}

.typeColors <- function(){
  c("8mer"="darkred","7mer-m8"="#44AA99","7mer-a1"="#117733", "7mer"="#117733",
    "6mer"="#4CAF50","6mer-m8"="#449d48", "6mer-a1"="#3c8c40",
    "g-bulged 8mer"="#5bb9e8", "g-bulged 7mer"="#88CCEE",
    "g-bulged 6mer"="#b5dff4", "wobbled 7mer"="grey85",
    "wobbled 8mer"="grey75", "non-canonical"="grey55", "+A"="darkred")
}


#' viewTargetAlignment
#'
#' @param m A GRanges of length 1 giving the information for a given match, as
#' produced by \code{\link{findSeedMatches}}.
#' @param miRNA A miRNA sequence, or a \code{\link{KdModel}} object of the miRNA
#' corresponding to the match in `m`; alternatively, a \code{\link{KdModelList}}
#' including the model.
#' @param seqs The sequences corresponding to the seqnames of `m`. Not needed if
#' `m` contains the target sequences.
#' @param flagBulgeMatches Logical; whether to flag matches inside the bulge
#' (default FALSE)
#' @param p3.params See \code{\link{findSeedMatches}}.
#' @param min3pMatch The minimum 3' alignment for any to be plotted
#' @param UGsub Logical; whether to show U-G matches
#' @param hideSingletons Logical; whether to hide isolated single base-pair
#' matches
#' @param outputType Either 'print' (default, prints to console), 'data.frame',
#' or 'plot'.
#' @param ... Passed to `text` if `outputType="plot"`.
#'
#' @return Returns nothing `outputType="print"`. If `outputType="data.frame"`,
#' returns a data.frame containing the alignment strings; if
#' `outputType="ggplot"` returns a `ggplot` object.
#' @importFrom stringi stri_reverse
#' @importFrom graphics par text
#' @export
#' @examples
#' data(SampleKdModel)
#' seq <- c(seq1="CGACCCCTATCACGTCCGCAGCATTAAAT")
#' m <- findSeedMatches(seq, SampleKdModel, verbose=FALSE)
#' viewTargetAlignment(m, miRNA=SampleKdModel, seqs=seq)
viewTargetAlignment <- function(m, miRNA, seqs=NULL, flagBulgeMatches=FALSE,
                                p3.params=list(), min3pMatch=3L,
                                hideSingletons=FALSE, UGsub=TRUE, ...,
                                outputType=c("print","data.frame",
                                             "plot","ggplot")){
  stopifnot(is(m,"GRanges"))
  stopifnot(length(m)==1)
  outputType <- match.arg(outputType)
  p3.params <- .check3pParams(p3.params)
  if(is.list(miRNA) && is(miRNA[[1]],"KdModelList") && !is.null(m$miRNA)){
    stopifnot(as.character(m$miRNA) %in% names(miRNA))
    miRNA <- miRNA[[as.character(m$miRNA)]]
  }
  mod <- miRNA
  if(is(miRNA,"KdModel")){
    miRNA <- miRNA$mirseq
  }else{
    stopifnot(is.character(miRNA) && length(miRNA)==1)
    if(nchar(miRNA)<15) stop("`miRNA` should be a full mature miRNA sequence")
  }
  if(is.null(seqs) && ( !("p3.mir.bulge" %in% colnames(mcols(m))) ||
                        !("sequence" %in% colnames(mcols(m))) ) ){
    stop("`m` does not contain target RNA sequences. Please provide them ",
           "via the `seqs` argument.")
  }
  if(!("p3.mir.bulge" %in% colnames(mcols(m)))){
    # re-scan to get additional data
    maxBulgeSize <- max(p3.params$maxMirLoop, p3.params$maxTargetLoop)
    internStart <- max(1L,start(m)-(nchar(miRNA)+maxBulgeSize-8L))
    seq2 <- subseq(DNAStringSet(seqs[as.character(seqnames(m))]),
                   internStart, end(m)+2L)
    tx_start <- start(m)
    m <- findSeedMatches(seq2, mod, keepMatchSeq=TRUE, p3.extra=TRUE,
                         p3.params=p3.params,
                         maxLogKd=0, minDist=-Inf, verbose=FALSE)
    m <- m[tx_start==start(m)+internStart-1L]
  }
  if(!("sequence" %in% colnames(mcols(m)))){
    stopifnot(as.character(seqnames(m)) %in% names(seqs))
    seqs <- DNAString(seqs[[as.character(seqnames(m))]])
    bulgeDiff <- max(m$p3.target.bulge-m$p3.mir.bulge,0L)
    r <- IRanges(max(start(m)-bulgeDiff-nchar(miRNA)+8L,1L), end(m)+2L)
    m$sequence <- as.character(unlist(extractAt(seqs, r)))
  }
  do3p <- TRUE
  if(m$p3.mir.bulge>p3.params$maxMirLoop |
     m$p3.target.bulge>p3.params$maxTargetLoop |
     abs(m$p3.mir.bulge-m$p3.target.bulge) > p3.params$maxLoopDiff ){
    m$p3.mir.bulge <- m$p3.target.bulge <- 3L
    do3p <- FALSE
  }
  mirseq <- gsub("T","U",miRNA)
  if(bulged <- grepl("g-bulged",m$type))
    mirseq <- paste0(substr(mirseq,1,5),"-",substr(mirseq,6,nchar(mirseq)))
  target <- stringi::stri_reverse(gsub("T","U",as.character(m$sequence)))
  target2 <- target
  mirseq2 <- as.character(complement(RNAString(mirseq)))
  mirseq2 <- paste0("A",substr(mirseq2, 2, nchar(mirseq2)))
  if(flagBulgeMatches){
    minBulge <- min(m$p3.mir.bulge,m$p3.target.bulge)
  }else{
    minBulge <- 0
  }
  mm <- .matchStrings(substr(mirseq2,1,8+bulged+minBulge),
                      substr(target,3,10+bulged+minBulge), UGsub=UGsub )
  # prevent wobble at first two positions:
  mm <- gsub("^(.)-","\\1-",gsub("^-"," ",mm))
  if(!flagBulgeMatches && m$p3.mir.bulge==m$p3.target.bulge){
    mm <- c(mm,rep(" ",m$p3.mir.bulge))
  }else if(m$p3.mir.bulge<m$p3.target.bulge){
    mm <- c(mm,rep(" ",m$p3.target.bulge-minBulge))
    di <- m$p3.target.bulge-m$p3.mir.bulge
    di1 <- paste(rep("-",ifelse(flagBulgeMatches,0,floor(di/2))),collapse="")
    di2 <- paste(rep("-",ifelse(flagBulgeMatches,di,ceiling(di/2))),collapse="")
    mirseq <- paste0(
      paste(substr(mirseq,1,8+bulged),collapse=""), di1,
      paste(substr(mirseq,9+bulged,8+bulged+m$p3.mir.bulge),collapse=""), di2,
      paste(substr(mirseq,9+bulged+m$p3.mir.bulge,nchar(mirseq)),collapse=""))
  }else{
    mm <- c(mm,rep(" ",m$p3.mir.bulge-minBulge))
    di <- m$p3.mir.bulge-m$p3.target.bulge
    target <- paste0(
      paste(substr(target,1,10),collapse=""),
      paste(rep("-",floor(di/2)),collapse=""),
      paste(substr(target,11,10+m$p3.target.bulge),collapse=""),
      paste(rep("-",ceiling(di/2)), collapse=""),
      paste(substr(target,11+m$p3.target.bulge,nchar(target)),collapse=""))
  }
  minl <- min(nchar(target),
              nchar(mirseq2)-8-m$p3.mir.bulge+10+m$p3.target.bulge)
  mirseq2 <- substr(mirseq2,9+m$p3.mir.bulge,nchar(mirseq2))
  target2 <- substr(target2,11+m$p3.target.bulge,nchar(target2))
  minl <- min(nchar(mirseq2),nchar(target2))
  mirseq2 <- substr(mirseq2,1,minl)
  target2 <- substr(target2,1,minl)
  mm2 <- .matchStrings(mirseq2, target2, UGsub)
  if(min3pMatch>1L &&
     !grepl(paste(rep("|",min3pMatch),collapse=""),
            gsub("-","|",paste(mm2,collapse=""),fixed=TRUE), fixed=TRUE)){
    mm2 <- rep(" ",sum(nchar(mm2)))
  }
  mm <- paste(c(mm,mm2),collapse="", sep="")
  sp <- function(x) paste0(rep(" ",x),collapse="")
  d <- data.frame(
    row.names=c("miRNA ",sp(6),"target"),
    "alignment"=c(paste0(sp(3),"3'-",stringi::stri_reverse(mirseq),"-5'",sp(5)),
                  paste0(sp(8),stringi::stri_reverse(mm),sp(8)),
                  paste0("5'-...",stringi::stri_reverse(target),"...-3'")))
  d$alignment <- paste0(vapply(max(nchar(d$alignment))-nchar(d$alignment),
                               FUN.VALUE=character(1),
                               FUN=function(x) paste0(rep(" ",x),collapse="")),
                        d$alignment)
  if(hideSingletons) d$alignment <- gsub(" | ", paste(rep(" ",3),collapse=""),
                                         d$alignment, fixed=TRUE)
  if(outputType=="data.frame") return(d)
  d2 <- paste0(paste(c("miRNA ","      ","target"), d$alignment), collapse="\n")
  if(outputType=="ggplot"){
    p <- ggplot(data.frame(x=1,y=1,label=d2), aes(x,y,label=label)) +
      theme_void() + geom_text(family="mono", fontface="bold")
    return(p)
  }else if(outputType=="plot"){
    par(xpd = NA, mar=c(0,0,0,0))
    plot(1, 1, ann=FALSE, bty='n', type='n', xaxt='n', yaxt='n')
    text(x=1, y=1, d2, family="mono", font=2, ...)
  }else{
    cat(paste0("\n",d2,"\n"))
  }
}

.matchStrings <- function(s1, s2, UGsub=TRUE){
  s1 <- strsplit(s1,"")[[1]]
  s2 <- strsplit(s2,"")[[1]]
  mm <- ifelse(s1==s2, "|", " ")
  if(UGsub){
    mm[s1!=s2 & ((s1=="A" & s2=="G") | (s1=="C" & s2=="U"))] <- "-"
  }
  mm <- gsub("\\-$", " ", paste(mm, collapse=""))
  mm <- gsub("\\-[^|]|[^|]\\-", "  ", mm)
  mm
}

#' Predicting and characterizing miRNA binding sites
#'
#' `findSeedMatches` takes a set of sequences and a set of miRNAs (given either
#' as target seeds, mature miRNA sequences, or a \code{\link{KdModelList}}).
#'
#' @param seqs A character vector or `DNAStringSet` of DNA sequences in which to
#' look.
#' @param seeds A character vector of 7-nt seeds to look for. If RNA, will be
#' reversed and complemented before matching. If DNA, they are assumed to be
#' the target sequence to look for. Alternatively, a list of objects of class
#' `KdModel` or an object of class `KdModelList` can be given.
#' @param shadow Integer giving the shadow, i.e. the number of nucleotides
#'  hidden at the beginning of the sequence (default 0).
#' @param onlyCanonical Logical; whether to restrict the search only to
#' canonical binding sites.
#' @param maxLogKd Maximum log_kd value to keep. This has a major impact on the
#' number of sites returned, and hence on the memory requirements. Set to Inf
#' to disable (_not_ recommended when running large scans!).
#' @param keepMatchSeq Logical; whether to keep the sequence (including flanking
#' dinucleotides) for each seed match (default FALSE).
#' @param minDist Integer specifying the minimum distance between matches of the
#' same miRNA (default 7). Closer matches will be reduced to the
#' highest-affinity. To disable the removal of overlapping features, use
#' `minDist=-Inf`.
#' @param p3.extra Logical; whether to keep extra information about 3'
#' alignment. Disable (default) this when running large scans, otherwise you
#' might hit your system's memory limits.
#' @param p3.params Named list of parameters for 3' alignment with slots
#' `maxMirLoop` (integer, default = 7), `maxTargetLoop` (integer, default = 9),
#' `maxLoopDiff` (integer, default = 4), `mismatch`
#' (logical, default = TRUE) and `GUwob` (logical, default = TRUE).
#' @param agg.params A named list with slots `a`, `b`, `c`, `p3`, `coef_utr`,
#' `coef_orf` and `keepSiteInfo` indicating the parameters for the aggregation.
#' Ignored if `ret!="aggregated"`. For further details see documentation of
#' `aggregateMatches`.
#' @param ret The type of data to return, either "GRanges" (default),
#' "data.frame", or "aggregated" (aggregates affinities/sites for each
#' seed-transcript pair).
#' @param BP Pass `BiocParallel::MulticoreParam(ncores, progressbar=TRUE)` to
#' enable multithreading.
#' @param verbose Logical; whether to print additional progress messages
#' (default on if not multithreading)
#' @param n_seeds Integer; the number of seeds that are processed in parallel to
#' avoid memory issues.
#' @param useTmpFiles Logical; whether to write results for single miRNAs in
#' temporary files (ignored when scanning for a single seed). Alternatively,
#' `useTmpFiles` can be a character vector of length 1 indicating the path to
#' the directory in which to write temporary files.
#' @param keepTmpFiles Logical; whether to keep the temporary files at the end
#' of the process; ignored if `useTmpFiles=FALSE`. Temporary files are removed
#' only upon successful completion of the function, meaning that they will not
#' be deleted in case of errors.
#'
#' @return A GRanges of all matches. If `seeds` is a `KdModel` or `KdModelList`,
#' the `log_kd` column will report the ln(Kd) multiplied by 1000, rounded and
#' saved as an integer. If `ret!="GRanges`, returns a data.frame.
#'
#' @importFrom BiocParallel bplapply SerialParam bpnworkers
#' @importFrom Seqinfo seqlevels
#' @import GenomicRanges
#' @importFrom Biostrings DNAString RNAString DNAStringSet subseq complement
#'             reverseComplement extractAt xscat padAndClip consensusMatrix
#' @importFrom pwalign pairwiseAlignment type pattern subject
#'             nucleotideSubstitutionMatrix
#' @importFrom S4Vectors mcols mcols<- metadata metadata<- Rle DataFrame head
#' @importFrom IRanges IRanges RleList
#' @export
#'
#' @examples
#' # we create mock RNA sequences and seeds:
#' seqs <- getRandomSeq(n=10)
#' seeds <- c("AAACCAC", "AAACCUU")
#' findSeedMatches(seqs, seeds)
findSeedMatches <- function( seqs, seeds, shadow=0L, onlyCanonical=FALSE,
                             maxLogKd=c(-1,-1.5), keepMatchSeq=FALSE,
                             minDist=7L, p3.extra=FALSE,
                             p3.params=list(maxMirLoop=7L, maxTargetLoop=9L,
                                            maxLoopDiff=4L, mismatch=TRUE,
                                            GUwob=TRUE),
                             agg.params=.defaultAggParams(),
                             ret=c("GRanges","data.frame","aggregated"),
                             BP=NULL, verbose=NULL, n_seeds=NULL,
                             useTmpFiles=FALSE, keepTmpFiles=FALSE){
  p3.params <- .check3pParams(p3.params)
  if(length(maxLogKd)==1) maxLogKd <- rep(maxLogKd,2)
  seqtype <- .guessSeqType(seqs)
  if(seqtype=="RNA")
    stop("Target sequences should be in DNA format.")
  if(is.character(seqs)) seqs <- DNAStringSet(seqs)
  mcols(seqs)$length <- length.seqs <- width(seqs)
  tx_info <- mcols(seqs)
  if(is.null(tx_info$ORF.length)){
    hasORF <- FALSE
    utr.length <- length.seqs
    orf.length <- 0L
    tx_info$ORF.length <- orf.length
    mcols(seqs)$ORF.length <- orf.length
    mcols(seqs)$C.length <- orf.length
  }else{
    hasORF <- TRUE
    orf.length <- tx_info$ORF.length
    utr.length <- ifelse(length.seqs > orf.length, length.seqs - orf.length, 0L)
    mcols(seqs)$ORF.length <- orf.length
    mcols(seqs)$C.length <- orf.length + shadow
    shadow <- 0L
  }
  tx_info$UTR.length <- utr.length
  length.info <- cbind(orf.length, utr.length)
  if(!is.null(names(seqs))) row.names(length.info) <- names(seqs)

  ret <- match.arg(ret)
  if(ret=="aggregated"){
    if(!is.list(agg.params)) agg.params <- as.list(agg.params)
    if(!all(names(agg.params) %in% names(.defaultAggParams())))
      stop("`agg.params` should be a named list with slots among ",
           "`a`, `b`, `c`, `p3`, `coef_utr`, `coef_orf` and `keepSiteInfo.")
    for(name in names(.defaultAggParams())){
      if(!name %in% names(agg.params)){
        agg.params[[name]] <- .defaultAggParams()[[name]]
      }
    }
  }
  seedInputType <- .checkSeedsInput(seeds)
  if(is.null(verbose)) {
    verbose <- is(seeds,"KdModel") || length(seeds)==1 || is.null(BP)
  }
  if(verbose) message("Preparing sequences...")
  maxLoop <- max(unlist(p3.params[c("maxMirLoop","maxTargetLoop")]))
  args <- .prepSeqs(seqs, seeds, shadow=shadow, pad=c(maxLoop+30L,8L))
  seqs <- args$seqs
  if("seeds" %in% names(args)) seeds <- args$seeds
  offset <- args$offset
  rm(args)

  params <- list(
    shadow=shadow,
    minDist=minDist,
    maxLogKd=maxLogKd,
    p3.params=p3.params
  )
  if(ret=="aggregated") params$agg.params <- agg.params

  if(is(seeds,"KdModel") || length(seeds)==1){
    useTmpFiles <- FALSE
    if(is.list(seeds[[1]])) seeds <- seeds[[1]]
    params$miRNA <- ifelse(is(seeds, "KdModel"), seeds$name, seeds)
    if(is.null(verbose)) verbose <- TRUE
    m <- .find1SeedMatches(seqs, seeds, keepMatchSeq=keepMatchSeq,
                           minDist=minDist, maxLogKd=maxLogKd,
                           onlyCanonical=onlyCanonical, p3.extra=p3.extra,
                           p3.params=p3.params, offset=offset,
                           verbose=verbose, ret=ret)
    if(length(m)==0) return(m)
    if(ret=="aggregated"){
      if(verbose) message("Aggregating...")
      ll <- as.data.frame(length.info)
      ll$transcript <- row.names(ll)
      m <- .aggregate_miRNA(m, ll, a=agg.params$a, b=agg.params$b,
                            c=agg.params$c, p3 = agg.params$p3,
                            coef_utr = agg.params$coef_utr,
                            coef_orf = agg.params$coef_orf,
                            keepSiteInfo = agg.params$keepSiteInfo, toInt=TRUE)
    }
  }else{
    if(is.null(BP)) BP <- SerialParam()
    if(is.null(verbose)) verbose <- !(bpnworkers(BP)>1 | length(seeds)>5)
    if(is.null(n_seeds)) n_seeds <- length(seeds)
    if(isTRUE(useTmpFiles) ||
       (is.character(useTmpFiles) && length(useTmpFiles)==1)){
      if(is.logical(useTmpFiles)){
        tmpDir <- tempdir()
      }else{
        tmpDir <- useTmpFiles
        useTmpFiles <- TRUE
      }
      stopifnot(!file.access(tmpDir, mode=2))
    }
    if(verbose) {
      if(is.numeric(BP$workers)) {
        message("Scanning with ", n_seeds, "seeds at a time on ", BP$workers,
                " cores...")
        if(useTmpFiles) message("Temporary writing results in:\n", tmpDir)
      } else {
        message("Scanning with ", n_seeds, " seeds at a time...")
      }
    }
    split_seeds <- split(seeds, ceiling(seq_along(seeds)/n_seeds))
    m <- lapply(split_seeds, function(seeds) {
      m <- bplapply(seeds, BPPARAM=BP, FUN=function(oneseed){
        if(ret=="aggregated" && is(oneseed, "KdModel") && agg.params$p3 == 0) {
          oneseed$mirseq <- NULL
        }
        m <- .find1SeedMatches(seqs=seqs, seed=oneseed,
                               keepMatchSeq=keepMatchSeq, minDist=minDist,
                               maxLogKd=maxLogKd, p3.extra=p3.extra,
                               onlyCanonical=onlyCanonical, p3.params=p3.params,
                               ret=ret, offset=offset, verbose=verbose)
        if(ret=="aggregated"){
          if(verbose) message("Aggregating...")
          if(length(m)==0) return(data.frame())
          ll <- as.data.frame(length.info)
          ll$transcript <- row.names(ll)
          m <- .aggregate_miRNA(m, ll, a=agg.params$a, b=agg.params$b,
                                c=agg.params$c, p3 = agg.params$p3,
                                coef_utr = agg.params$coef_utr,
                                coef_orf = agg.params$coef_orf,
                                keepSiteInfo = agg.params$keepSiteInfo,
                                toInt=TRUE)
        }
        if(!useTmpFiles) return(m)
        f <- paste0(ifelse(is.character(oneseed), oneseed, oneseed$name), "_")
        f <- tempfile(f, tmpdir=tmpDir)
        saveRDS(m, f, compress=FALSE)
        rm(m)
        gc(verbose = FALSE)
        f
      })
    })
    m <- do.call(c, m)
    names(m) <- NULL

    if(!is.character(seeds)) {
      seeds <- vapply(seeds, FUN=function(x){
        if(is.null(x$name)) return(x$canonical.seed)
        x$name
      }, character(1))
    }
    names(m) <- seeds

    if(useTmpFiles){
      if(verbose) message("Scan completed, reading and aggregating results ",
                          "from temporary files...")
      ff <- unlist(m,use.names=FALSE)
      m <- lapply(m, FUN=function(x) readRDS(x))
    }

    if(ret=="GRanges"){
      m <- .unlistGRL(m, .id="miRNA")
      names(m) <- row.names(m) <- NULL
      mcols(m)$miRNA <- Rle(as.factor(mcols(m)$miRNA))
    }else{
      m <- lapply(m, as.data.frame)
      m <- data.table::rbindlist(m, fill=TRUE, use.names=TRUE,
                                               idcol="miRNA")
      for(f in colnames(m)){
        if(is.numeric(m[[f]])) m[[f]][is.na(m[[f]])] <- 0L
      }
      m$miRNA <- as.factor(m$miRNA)
      row.names(m) <- NULL
      m$transcript <- as.factor(m$transcript)
    }
  }
  if(ret=="GRanges"){
    metadata(m)$call.params <- params
    metadata(m)$tx_info <- as.data.frame(tx_info)
  }else{
    m <- as.data.frame(m)
    attr(m, "call.params") <- params
    attr(m, "tx_info") <- as.data.frame(tx_info)
  }
  if(ret=="aggregated" && onlyCanonical) m[["non-canonical"]] <- NULL
  if(useTmpFiles && !keepTmpFiles) unlink(ff)
  return(m)
}

# scan for a single seed
.find1SeedMatches <- function(seqs, seed, keepMatchSeq=FALSE, maxLogKd=-1,
                              minDist=1L, onlyCanonical=FALSE, p3.extra=FALSE,
                              p3.params=list(), offset=0L,
                              ret=c("GRanges","data.frame","aggregated"),
                              verbose=FALSE){
  ret <- match.arg(ret)
  p3.params <- .check3pParams(p3.params)

  if(is.null(maxLogKd)) maxLogKd <- c(Inf,Inf)
  if(length(maxLogKd)==1) maxLogKd <- rep(maxLogKd,2)

  if(verbose) message("Scanning for matches...")

  if(isPureSeed <- is.character(seed)){
    stopifnot(nchar(seed)>6)
    if(nchar(seed) %in% 7:8){
      mirseq <- NULL
      if(verbose && p3.extra)
        warning("`p3.extra` ignored when input is only a seed")
      p3.extra <- FALSE
    }else{
      mirseq <- gsub("U","T",seed)
      seed <- as.character(reverseComplement(DNAStringSet(substr(mirseq,2,8))))
      seed <- paste0(seed,"A")
      if(verbose && (nchar(mirseq)<18 | nchar(mirseq)>24))
        warning("The `seed` given seems to be neither a miRNA target seed nor",
                "a mature miRNA sequence! Scanning for ", seed)
    }
    pat <- substr(seed,2,7)
    if(!onlyCanonical)
      pat <- paste0(pat,"|",substr(seed,2,3),"G",substr(seed,4,7))
    pos <- gregexpr(paste0("(?=.",pat,".)"), seqs, perl=TRUE)
  }else{
    mod <- seed
    seed <- mod$canonical.seed
    mirseq <- mod$mirseq
    if(onlyCanonical){
      patt <- paste0(".",substr(seed,2,7),".")
    }else{
      patt <- .build4mersRegEx(seed)
    }
    pos <- gregexpr(paste0("(?=",patt,")"), seqs, perl=TRUE)
  }
  pos <- lapply(lapply(pos, as.integer), y=-1L, setdiff)
  if(sum(lengths(pos))==0){
    if(verbose) message("Nothing found!")
    return(GRanges())
  }
  m <- GRanges(rep(names(seqs), lengths(pos)),
               IRanges(start=unlist(pos), width=8))
  rm(pos)
  m <- m[order(seqnames(m))]

  if(verbose) message("Extracting sequences and characterizing matches...")
  r <- ranges(m)
  if(isPureSeed){
    r <- split(r, seqnames(m))
    names(r) <- NULL
    ms <- as.factor(unlist(extractAt(seqs[seqlevels(m)], r))) # 8mers
    mcols(m)$type <- getMatchTypes(as.factor(ms), substr(seed,1,7))
    if(keepMatchSeq && !p3.extra) mcols(m)$sequence <- ms
    m <- m[order(seqnames(m), m$type)]
  }else{
    start(r) <- start(r)-2L
    end(r) <- end(r)+2L
    r <- split(r, seqnames(m))
    names(r) <- NULL
    ms <- unlist(extractAt(seqs[seqlevels(m)], r))  # 12mers
    mcols(m) <- cbind(mcols(m), assignKdType(ms, mod))
    if(keepMatchSeq) mcols(m)$sequence <- ms
    if(maxLogKd[[1]]!=Inf){
      if(all(maxLogKd>=0)) maxLogKd <- -maxLogKd
      if(all(maxLogKd > -10)) maxLogKd <- maxLogKd*1000L
      m <- m[which(m$log_kd <= as.integer(round(maxLogKd[1])))]
    }else{
      m <- m[!is.na(m$log_kd)]
    }
    m <- m[order(seqnames(m), m$log_kd, m$type)]
  }
  rm(ms)
  if(minDist>-Inf){
    if(verbose) message("Removing overlaps...")
    m <- removeOverlappingRanges(m, minDist=minDist, ignore.strand=TRUE)
  }
  if(!is.null(mirseq)){
    if(verbose) message("Performing 3' alignment...")
    maxLoop <- max(unlist(p3.params[c("maxMirLoop","maxTargetLoop")]))
    plen <- maxLoop+nchar(mirseq)-8L
    r <- ranges(m)
    start(r) <- start(r)-plen-1L
    end(r) <- start(r)+plen
    if(isPureSeed && keepMatchSeq && p3.extra) end(r) <- end(r) + 10L
    r <- split(r, seqnames(m))
    names(r) <- NULL
    ms <- unlist(extractAt(seqs[seqlevels(m)], r))
    rm(r)
    names(ms) <- NULL
    p3 <- get3pAlignment( subseq(ms,1,plen+1L), mirseq, siteType=mcols(m)$type,
                          allow.mismatch=p3.params$mismatch,
                          maxMirLoop=p3.params$maxMirLoop,
                          maxLoopDiff=p3.params$maxLoopDiff,
                          maxTargetLoop=p3.params$maxTargetLoop,
                          TGsub = p3.params$GUwob)
    if(p3.extra){
      mcols(m) <- cbind(mcols(m), p3)
      if(keepMatchSeq){
        if(isPureSeed){
          mcols(m)$sequence <- ms
        }else{
          mcols(m)$sequence <- xscat(ms, subseq(mcols(m)$sequence,3L))
        }
      }
    }else{
      mcols(m)$p3.score <- p3$p3.score
      mcols(m)$note <- p3$note
    }
    rm(ms)
    mcols(m)$note <- Rle(mcols(m)$note)
  }
  if(!is.null(mcols(seqs)$C.length) && !all(mcols(seqs)$ORF.length == 0)) {
    mcols(m)$ORF <-
      start(m) <= mcols(seqs[seqlevels(m)])[as.integer(seqnames(m)),"C.length"]
    if(!isPureSeed && maxLogKd[2]!=Inf && maxLogKd[2]!=maxLogKd[1]){
      m <- m[which(!m$ORF | m$log_kd <= as.integer(round(maxLogKd[2])))]
    }
    mcols(m)$ORF <- Rle(mcols(m)$ORF)
  }
  names(m) <- NULL
  if(offset!=0) m <- IRanges::shift(m, -offset)
  if(ret=="data.frame") m <- .gr2matchTable(m)
  return(m)
}

.checkSeedsInput <- function(seeds){
  if(is.character(seeds)){
    if(all(nchar(seeds) %in% c(7,8))) return("seed")
    if(all(nchar(seeds) >= 16 & nchar(seeds) <= 26)) return("mirseq")
    stop("If `seeds` is a character vector, it should either be vector of ",
         "seeds (corresponding DNA sequence, 7 or 8 nucleotides), or of mature",
         " miRNA sequences (RNA sequence).\n",
         "The size of the characters in `seeds` is compatible with neither.")
  }
  if(!is(seeds,"KdModel") && !is(seeds,"KdModelList") &&
     !(is.list(seeds) && all(vapply(seeds, class, character(1))=="KdModel")))
    stop("`seeds` should either be a character vector or an object of class ",
         "`KdModel` or `KdModelList`")
  NULL
}


.gr2matchTable <- function(m, include_name=FALSE, include_ORF=TRUE, p3=TRUE){
  d <- DataFrame(transcript=as.factor(seqnames(m)), start=start(m))
  if(include_name) d$miRNA <- as.factor(m$miRNA)
  if(include_ORF && !is.null(m$ORF)) d$ORF <- m$ORF
  d$type <- m$type
  d$log_kd <- m$log_kd
  d$note <- m$note
  for(f in c("p3.mir.bulge",
             "p3.target.bulge",
             "p3.mismatch",
             "p3.matches",
             "p3.score")){
    if(p3 && f %in% colnames(mcols(m)))
      d[[f]] <- mcols(m)[[f]]
  }
  d
}

.check3pParams <- function(p3.params){
  stopifnot(is.list(p3.params))
  def <- list(maxMirLoop=7L, maxTargetLoop=9L, maxLoopDiff=4L, mismatch=TRUE, GUwob=TRUE)
  for(f in names(def)){
    if(!(f %in% names(p3.params))) p3.params[[f]] <- def[[f]]
  }
  stopifnot(is.logical(p3.params$mismatch))
  stopifnot(is.logical(p3.params$GUwob))
  stopifnot(is.numeric(unlist(p3.params[names(def)[seq_len(3)]])))
  for(f in names(def)[seq_len(3)]) p3.params[[f]] <- as.integer(p3.params[[f]])
  p3.params
}

#' Finds 3' complementary binding of a miRNA
#'
#' Performs a local alignment of the miRNA 3' sequence (determined by
#' `mir3p.start`) on given the given sequences.
#'
#' @param seqs A set of sequences in which to look for 3' matches (i.e. upstream
#' of the seed match)
#' @param mirseq The sequence of the mature miRNA
#' @param siteType The optional type of seed-complementarity, as returned by
#' \code{\link{getMatchTypes}}. This is needed to identify slicing/TDMD sites.
#' If given, should be a vector of the same length as `seqs`.
#' @param mir3p.start The position in `mirseq` in which to start looking
#' @param allow.mismatch Logical; whether to allow mismatches
#' @param TGsub Logical; whether to allow T/G substitutions.
#' @param maxMirLoop Maximum miRNA loop size
#' @param maxTargetLoop Maximum target loop size
#' @param maxLoopDiff Maximum size difference between miRNA and target loops
#'
#' @return A data.frame with one row for each element of `seqs`, indicating the
#' size of the miRNA bulge, the size of the target mRNA bulge, the number of
#' mismatches at the 3' end, and the partial 3' alignment score (i.e. roughly
#' the number of consecutive matching nucleotides)
#'
#' @export
#'
#' @examples
#' get3pAlignment(seqs="NNAGTGTGCCATNN", mirseq="TGGAGTGTGACAATGGTGTTTG")
get3pAlignment <- function(seqs, mirseq, mir3p.start=9L, allow.mismatch=TRUE,
                           maxMirLoop=7L, maxTargetLoop=9L, maxLoopDiff=4L,
                           TGsub=TRUE, siteType=NULL){
  if(!is.null(siteType)) stopifnot(length(seqs)==length(siteType))
  mir.3p <- as.character(DNAString(substr(x=mirseq,
                                          start=mir3p.start,
                                          stop=nchar(mirseq))))

  if(!is(seqs, "XStringSet") && !is(seqs, "XString")) seqs <- DNAString(seqs)
  seqs <- reverseComplement(seqs)
  subm <- .default3pSubMatrix(ifelse(allow.mismatch,-3L,-Inf), TG=TGsub)
  mcols(seqs) <- NULL
  al <- pairwiseAlignment(seqs, mir.3p, type="local", substitutionMatrix=subm)
  df <- data.frame( p3.mir.bulge=start(subject(al))-1L,
                    p3.target.bulge=start(pattern(al))-1L )
  df$p3.mismatch <- nchar(mir.3p)-width(pattern(al))-df$p3.mir.bulge
  df$p3.score <- as.integer(score(al))
  diff <- abs(df$p3.mir.bulge-df$p3.target.bulge)
  df$p3.score <- ifelse(diff > 2L, df$p3.score - (diff - 2L), df$p3.score)
  df[which(df$p3.mir.bulge>maxMirLoop | df$p3.target.bulge>maxTargetLoop |
             diff>maxLoopDiff),
     c("p3.mir.bulge","p3.target.bulge","p3.score")] <- 0L
  if(!is.integer(df$p3.score)) df$p3.score <- as.integer(round(df$p3.score))
  if(!is.null(siteType)){
    df$type <- siteType
    df$note <- .TDMD(df, mirseq, acceptWobble=TGsub)
    n_9_11 <- substr(as.character(mirseq),9,11)
    if(length(w <- which(df$note %in% c("Slicing","Slicing?")))>0 &&
       sum(w2 <- !grepl(paste0("^",n_9_11), as.character(seqs)[w]))>0){
      # for slicing sites, ensure that positions 9-11 are complementary
      df$note[w[which(w2)]] <- "-"
    }
    df$type <- NULL
  }
  df
}

.TDMD <- function(m, mirseq, acceptWobble=TRUE){
  tt <- c("8mer","7mer-m8","7mer-a1","g-bulged 8mer")
  if(acceptWobble) tt <- c(tt, "wobbled 8mer","wobbled 7mer")
  is78 <- which(m$type %in% tt)
  m2 <- m[is78,]
  isWobble <- m2$type %in% c("wobbled 8mer","wobbled 7mer")
  isBulged <- m2$type=="g-bulged 8mer"
  m2$TDMD <- rep(1L,nrow(m2))
  absbulgediff <- abs(m2$p3.mir.bulge-m2$p3.target.bulge)
  w <- which(m2$p3.mismatch<=1L & m2$p3.mir.bulge<=7L & !isBulged & !isWobble &
             m2$p3.mir.bulge > 1L & absbulgediff <= 4L & m2$p3.score >= 6L)
  if(length(w)>0) m2$TDMD[w] <- 2L
  w <- which( m2$type != "7mer-a1"  &
              m2$p3.score >= 7L & m2$p3.mir.bulge == 0L & absbulgediff == 0L )
  if(length(w)>0) m2$TDMD[w] <- 3L
  m2$not.bound <- nchar(mirseq) - 8L - m2$p3.score
  w <- which(m2$TDMD == 3L & m2$p3.mismatch <= 1L & m2$not.bound <= 1L &
               !isWobble & !isBulged)
  if(length(w)>0) m2$TDMD[w] <- 4L
  TDMD <- rep(1L,nrow(m))
  TDMD[is78] <- m2$TDMD
  factor(TDMD, levels = 1L:4L, labels = c("-",
                                          "TDMD?",
                                          "Slicing?",
                                          "Slicing"))
}

.default3pSubMatrix <- function(mismatch=-3, TG=TRUE){
  subm <- nucleotideSubstitutionMatrix(match=1, mismatch=mismatch)
  l <- c("A","C","G","T","N")
  subm <- subm[l,l]
  if(TG) subm["A", "G"] <- subm["C", "T"] <- 0
  subm
}


#' removeOverlappingRanges
#'
#' Removes elements from a GRanges that overlap (or are within a given distance
#' of) other elements higher up in the list (i.e. assumes that the ranges are
#' sorted in order of priority). The function handles overlaps between more than
#' two ranges by successively removing those that overlap higher-priority ones.
#'
#' @param x A GRanges, sorted by (decreasing) importance.
#' @param minDist Minimum distance between ranges.
#' @param retIndices Logical; whether to return the indices of entries to
#' remove, rather than the filtered GRanges.
#' @param ignore.strand Logical. Whether the strand of the input ranges should
#' be ignored or not.
#'
#' @return A filtered GRanges, or an integer vector of indices to be removed if
#' `retIndices==TRUE`.
#' @export
#' @import GenomicRanges
#' @examples
#' library(GenomicRanges)
#' gr <- GRanges(seqnames=rep("A",4), IRanges(start=c(10,25,45,35), width=6))
#' removeOverlappingRanges(gr, minDist=7)
removeOverlappingRanges <- function(x, minDist=7L, retIndices=FALSE,
                                    ignore.strand=FALSE){
  red <- GenomicRanges::reduce(x, with.revmap=TRUE, min.gapwidth=minDist,
                               ignore.strand=ignore.strand)$revmap
  red <- red[lengths(red)>1]
  if(length(red)==0){
    if(retIndices) return(c())
    return(x)
  }
  i <- seq_along(x)
  toRemove <- c()
  while(length(red)>0){
    ## for each overlap set, we flag the index (relative to i) of the maximum
    ## (i.e. lowest in the list)
    top <- min(red) ## indexes of the top entry per overlap set, relative to i
    ## overlap of non-top entries to the top entries:
    o <- IRanges::overlapsAny(x[i[-top]],x[i[top]],maxgap=minDist)
    torem <- i[-top][which(o)] ## entries to remove, relative to x
    toRemove <- c(toRemove, torem) ## relative to x
    i <- setdiff(i,torem)
    ## and check again overlaps among this subset (revmap ind are relative to i)
    red <- GenomicRanges::reduce(x[i], with.revmap=TRUE, min.gapwidth=minDist,
                                 ignore.strand=ignore.strand)$revmap
    red <- red[lengths(red)>1]
  }
  if(retIndices) return(toRemove)
  if(length(toRemove)>0) x <- x[-toRemove]
  x
}

## determines target and seed sequence type, converts if necessary,
## and adds padding/shadow
.prepSeqs <- function(seqs, seeds, shadow=0, pad=c(0,0)){
  if(is.null(names(seqs))) names(seqs) <- paste0("seq",seq_along(seqs))
  seqtype <- .guessSeqType(seqs)
  if(seqtype=="RNA")
    stop("Both the seeds and the target sequences should be in DNA format.")
  ret <- list()
  if( is(seeds, "KdModel") ||
      (is.list(seeds) && all(vapply(seeds, is.list, logical(1)))) ){
    if(is.null(names(seeds)))
      stop("If `seeds` is a list of KdModels, it should be named.")
  }else{
    if(is.null(names(seeds))) n <- names(seeds) <- seeds
    seeds <- gsub("U", "T", seeds)
    names(seeds) <- n
    ret$seeds <- seeds
  }
  seqnms <- names(seqs)
  if(is.character(seqs)) seqs <- DNAStringSet(seqs)
  names(seqs) <- seqnms
  shadow <- max(c(0,shadow-1))
  ret$offset <- max(c(0,pad[1]-max(0,shadow)))
  seqs <- seqs[lengths(seqs)>=(shadow+8)]
  seqs <- subseq(seqs,1+shadow,lengths(seqs))
  seqs <- padAndClip(seqs, views=IRanges(start=1-shadow-ret$offset,
                                         width=lengths(seqs)+
                                           shadow+ret$offset+pad[2]),
                     Lpadding.letter = "N", Rpadding.letter = "N")
  if(!is.null(mcols(seqs)$C.length))
    mcols(seqs)$C.length <- mcols(seqs)$C.length + ret$offset + shadow
  c(ret, list(seqs=seqs))
}

#' getMatchTypes
#'
#' Given a seed and a set of sequences matching it, returns the type of match.
#'
#' @param x A character vector of short sequences.
#' @param seed A 7 or 8 nucleotides string indicating the seed (5' to 3'
#' sequence of the target RNA). If of length 7, an "A" will be appended.
#' @param checkWobble Whether to flag wobbled sites
#'
#' @return A factor of match types.
#' @export
#'
#' @examples
#' x <- c("AACACTCCAG","GACACTCCGC","GTACTCCAT","ACGTACGTAC")
#' getMatchTypes(x, seed="ACACTCCA")
getMatchTypes <- function(x, seed, checkWobble=TRUE){
  if(is.factor(x)) return(getMatchTypes(levels(x), seed)[as.integer(x)])
  x <- as.character(x)
  y <- rep(1L,length(x))
  seed <- as.character(seed)
  stopifnot(length(seed)==1)
  stopifnot(nchar(seed) %in% 7:8)
  if(nchar(seed)==7) seed <- paste0(seed,"A")
  seed6 <- substr(seed,2,7)
  seedGb6 <- paste0(substr(seed,2,3),"G",substr(seed,4,7))
  seedGb7 <- paste0(substr(seed,1,3),"G",substr(seed,4,7))
  y[grep(paste0("[ACGTN][ACGTN]",substr(seed,3,8)),x)] <- 2L # 6mer-a1
  y[grep(paste0(substr(seed,1,6),"[ACGTN][ACGTN]"),x)] <- 3L # 6mer-m8
  if(substr(seedGb6,2,7)!=seed6)
    y[grep(seedGb6,x,fixed=TRUE)] <- 4L # g-bulged 6mer
  y[grep(paste0("[ACGTN]",seed6),x)] <- 5L # 6mer
  if(seedGb6!=substr(seed,1,7)){
    y[grep(paste0(seedGb6,"A|",seedGb7),x)] <- 6L # g-bulged 7mer
    y[grep(paste0(seedGb7,"A"),x,fixed=TRUE)] <- 7L # g-bulged 8mer
  }
  if(checkWobble){
    y[.isWobble(x,seed)] <- 8L # wobbled 7-mer
    y[.isWobble(x,seed,FALSE)] <- 9L # wobbled 8-mer
  }
  y[grep(paste0("[ACGTN]",substr(seed,2,8)),x)] <- 10L # 7mer-a1
  y[grep(substr(seed,1,7),x,fixed=TRUE)] <- 11L # 7mer-m8
  y[grep(seed,x,fixed=TRUE)] <- 12L # 8mer
  factor(y, levels=12L:1L, labels=.matchLevels())
}

.isWobble <- function(x, seed, allow7mer=TRUE, positions=2:5){
  seed <- strsplit(seed,"")[[1]]
  if(allow7mer) seed <- head(seed,7)
  wo <- c(A="G", C="T")
  seeds <- vapply(intersect(which(seed %in% names(wo)), positions),
                  FUN.VALUE=character(1), FUN=function(i){
    seed[i] <- wo[seed[i]]
    paste(seed, collapse="")
  })
  if(length(seeds)==0) return(rep(FALSE, length(x)))
  grepl(paste(seeds, collapse="|"), x)
}

.matchLevels <- function(withA=TRUE){
  if(withA) return(c("8mer","7mer-m8","7mer-a1","wobbled 8mer","wobbled 7mer",
                     "g-bulged 8mer","g-bulged 7mer","6mer","g-bulged 6mer",
                     "6mer-m8","6mer-a1","non-canonical"))
  c("7mer","7mer","6mer","wobbled 8mer","wobbled 7mer","g-bulged 7mer",
    "g-bulged 6mer","6mer","g-bulged 6mer","6mer-m8","non-canonical",
    "non-canonical")
}

.defaultAggParams <- function(){
  list(a=0.007726,
    b=0.5735,
    c=0.1810,
    p3=0.051,
    coef_utr = -0.17106,
    coef_orf = -0.21546,
    keepSiteInfo = TRUE
    )
}

#' @importFrom IRanges FactorList
.unlistGRL <- function(m, .id=NULL, tryStandard=TRUE, verbose=FALSE){
  # to avoid c-stack errors on some systems
  if(tryStandard){
    gr <- try(unlist(GRangesList(m)), silent=TRUE)
    if(!is(gr,"try-error")){
      if(!is.null(.id))
        mcols(gr)[[.id]] <- rep(as.factor(names(m)), lengths(m))
      return(gr)
    }
  }
  if(verbose) message("Ranges...")
  gr <- GRanges(unlist(FactorList(lapply(m,seqnames)), use.names=FALSE),
                IRanges(unlist(lapply(m, start)), unlist(lapply(m,end))))
  for(f in colnames(mcols(m[[1]]))){
    if(verbose) message(f)
    x <- lapply(m, FUN=function(x) mcols(x)[[f]])
    if(is(x[[1]],"Rle")){
      x <- unlist(RleList(x), use.names=FALSE)
    }else if(is.factor(x[[1]])){
      x <- unlist(FactorList(x), use.names=FALSE)
    }else{
      x <- unlist(x)
    }
    mcols(gr)[[f]] <- x
  }
  if(verbose) message("Names...")
  if(!is.null(.id))
    mcols(gr)[[.id]] <- rep(as.factor(names(m)), lengths(m))
  gr
}

data("SampleKdModel", package="scanMiR")
seq1 <- DNAStringSet(c(seq = paste(c(rep("N", 20), "AGCATTAA", rep("N", 20)),
                                   collapse = "")))

gr <- findSeedMatches(seq1, SampleKdModel, verbose = FALSE, ret = "GRanges")
df <- findSeedMatches(seq1, SampleKdModel, verbose = FALSE, ret = "data.frame")
agg <- findSeedMatches(seq1, SampleKdModel, verbose = FALSE, ret = "aggregated")

test_that("Scan returns the right format", {
  expect_s4_class(gr, "GRanges")
  expect_s3_class(df, "data.frame")
  expect_s3_class(agg, "data.frame")
})

test_that("Scan returns the right match", {
  expect_identical(as.character(seqnames(gr)), "seq")
  expect_identical(ranges(gr), IRanges(21, 28))
  expect_identical(as.character(gr$type), "8mer")
})

seq2 <- DNAStringSet(c(seq = paste(c(rep("N", 5), "AGCATTAA", rep("N", 7),
                                     "AGCATTAA", rep("N", 20)), collapse = "")))
gr_s0 <- findSeedMatches(seq2, SampleKdModel, verbose = FALSE, shadow = 0L)
gr_s15 <- findSeedMatches(seq2, SampleKdModel, verbose = FALSE, shadow = 15L)

test_that("Shadow works", {
  expect_identical(length(gr_s0), 2L)
  expect_identical(length(gr_s15), 1L)
})

seq_nc <- DNAStringSet(c(seq = paste(c(rep("N", 5), "AGTATTAA", rep("N", 7),
                                       "AGCATTAA", rep("N", 20)),
                                     collapse = "")))
gr_withNC <- findSeedMatches(seq_nc, SampleKdModel, verbose = FALSE,
                             onlyCanonical = FALSE)
gr_withoutNC <- findSeedMatches(seq_nc, SampleKdModel, verbose = FALSE,
                                onlyCanonical = TRUE)

test_that("OnlyCanonical works", {
  expect_identical(length(gr_withNC), 2L)
  expect_identical(length(gr_withoutNC), 1L)
})

gr_ms <- findSeedMatches(seq1, SampleKdModel, verbose = FALSE, ret = "GRanges",
                         keepMatchSeq = TRUE)
test_that("KeepMatchSeq works", {
  expect_identical(as.character(gr_ms$sequence$seq), "NNAGCATTAANN")
})

gr_md_0 <- findSeedMatches(seq_nc, SampleKdModel, verbose = FALSE,
                           onlyCanonical = FALSE, minDist = 0L)
gr_md_10 <- findSeedMatches(seq_nc, SampleKdModel, verbose = FALSE,
                            onlyCanonical = FALSE, minDist = 10L)
test_that("minDist works", {
  expect_identical(length(gr_md_0), 2L)
  expect_identical(length(gr_md_10), 1L)
  expect_identical(as.character(gr_md_10$type), "8mer")
})

char_seq <- setNames(as.character(seq1$seq), "seq")
gr_char <- findSeedMatches(char_seq, SampleKdModel, verbose = FALSE,
                           ret = "GRanges")
test_that("Scan works for character sequence", {
  expect_identical(gr_char, gr)
})

char_seed <- SampleKdModel$canonical.seed
gr_char_seed <- findSeedMatches(seq1, char_seed, verbose = FALSE,
                                ret = "GRanges")
test_that("Scan works for character seed", {
  expect_identical(ranges(gr), ranges(gr_char_seed))
  expect_identical(gr$type, gr_char_seed$type)
})

test_that("Aggregation works", {
  expect_identical(nrow(agg), 1L)
  expect_identical(names(agg)[2], "repression")
})

seq_ORF <- seq2
mcols(seq_ORF)$ORF.length <- 15
gr_ORF <- findSeedMatches(seq_ORF, SampleKdModel, verbose = FALSE)
df_ORF <- findSeedMatches(seq_ORF, SampleKdModel, verbose = FALSE,
                          ret = "data.frame")
agg_ORF <- findSeedMatches(seq_ORF, SampleKdModel, verbose = FALSE,
                          ret = "aggregated")
test_that("Scanning ORF region works", {
  expect_identical(as.logical(gr_ORF$ORF), c(TRUE, FALSE))
  expect_s3_class(metadata(gr_ORF)$tx_info, "data.frame")
  expect_identical(names(metadata(gr_ORF)$tx_info), c("ORF.length", "length",
                                                      "UTR.length"))
  expect_identical(as.numeric(metadata(gr_ORF)$tx_info), c(15,48,33))
  expect_identical(df_ORF$ORF, c(TRUE, FALSE))
  expect_identical(attr(df_ORF, "tx_info"), metadata(gr_ORF)$tx_info)
  expect_identical(attr(agg_ORF, "tx_info"), metadata(gr_ORF)$tx_info)
  expect_identical(agg_ORF$ORF.canonical, 1L)
})

x <- c("AACACTCCAG","GACACTCCGC","GTACTCCAT","ACGTACGTAC")
matchTypes <- as.character(getMatchTypes(x, seed="ACACTCCA"))
test_that("Correct match types are returned", {
  expect_identical(matchTypes, c("8mer", "7mer-m8", "6mer-a1", "non-canonical"))
})

gr_OR_0 <- removeOverlappingRanges(gr_withNC, minDist = 0L)
gr_OR_10 <- removeOverlappingRanges(gr_withNC, minDist = 10L)
test_that("Correctly removes overlapping ranges", {
  expect_identical(length(gr_OR_0), 2L)
  expect_identical(length(gr_OR_10), 1L)
  expect_identical(as.character(gr_OR_10$type), "8mer")
})

test_that("3p alignment works correctly", {
  al3p <- get3pAlignment(seqs="TGTTTAAGAACAACAAAATCACCAT",
                         mirseq="TGGAAGACTAGTGATTTTGTTGTT", siteType="8mer")
  expect_equal(as.numeric(al3p[1,1:4]), c(2,3,0,14))
  expect_identical(as.character(al3p$note), "TDMD?")
  al3p <- get3pAlignment( "GCTAAGTTCTCCCAACAACATGAA", "TAGGTAGTTTCATGTTGTTGGG",
                          siteType="8mer")
  expect_equal(as.numeric(al3p[1,1:4]), c(0,0,0,14))
  expect_identical(as.character(al3p$note), "Slicing")
})

test_that("viewTargetAlignment works", {
  expect_s3_class(viewTargetAlignment(gr_OR_0[1], SampleKdModel, seq_nc,
                                      outputType="ggplot"), "ggplot")
  al <- viewTargetAlignment(gr_OR_0[1], SampleKdModel, seq_nc,
                            outputType="data.frame")
  expect_identical(al$alignment[2],
                   "                   |||-|        ||||||||        ")
})

## ----include=FALSE------------------------------------------------------------
library(BiocStyle)

## ----echo = FALSE, out.width="35%", fig.align = 'right'-----------------------
knitr::include_graphics(system.file('docs', 'sticker.svg', package = 'scanMiR'))

## -----------------------------------------------------------------------------
library(scanMiR)

# seed sequence of hsa-miR-155-5p
seed <- "AGCAUUAA"

# load a sample transcript
data("SampleTranscript")

# run scan
matches <- findSeedMatches(SampleTranscript, seed, verbose = FALSE)
matches

## -----------------------------------------------------------------------------
# full sequence of the mature miR-155-5p transcript
miRNA <- "UUAAUGCUAAUCGUGAUAGGGGUU"

# run scan
matches <- findSeedMatches(SampleTranscript, miRNA, verbose = FALSE)
matches

## -----------------------------------------------------------------------------
viewTargetAlignment(matches[1], miRNA, SampleTranscript)

## -----------------------------------------------------------------------------
# load sample KdModel
data("SampleKdModel")

# run scan
matches <- findSeedMatches(SampleTranscript, SampleKdModel, verbose = FALSE)
matches

## ----echo = FALSE, out.width="80%", fig.align = 'right', fig.cap="Adapted from Grimson et al. 2007"----
knitr::include_graphics(system.file('docs', 'sitetypes.jpg', package = 'scanMiR'))

## ----message = FALSE----------------------------------------------------------
library(Biostrings)

# generate set of random sequences
seqs <- DNAStringSet(getRandomSeq(length = 1000, n = 10))

# add vector of ORF lengths
mcols(seqs)$ORF.length <- sample(500:800, length(seqs))

# run scan
matches2 <- findSeedMatches(seqs, SampleKdModel, verbose = FALSE)
head(matches2)

## -----------------------------------------------------------------------------
viewTargetAlignment(matches[1], SampleKdModel, SampleTranscript)

## -----------------------------------------------------------------------------
agg_matches <- aggregateMatches(matches2)
head(agg_matches)

## -----------------------------------------------------------------------------
unlist(scanMiR:::.defaultAggParams())

## ----sessionInfo, echo=FALSE--------------------------------------------------
sessionInfo()

#' @rdname SampleTranscript
#' @name SampleTranscript
#' @aliases SampleTranscript
#'
#' @title Example transcript sequence
#'
#' @description
#' An artificial transcript sequence used for examples.
#'
#' @return a named character vector of length 1.
NULL

#' @rdname SampleKdModel
#' @name SampleKdModel
#' @aliases SampleKdModel
#'
#' @title Example KdModel (hsa-miR-155-5p)
#'
#' @description
#' `KdModel` for hsa-miR-155-5p, based on Kd predictions from the CNN of
#' [McGeary, Lin et al. (2019)](https://dx.doi.org/10.1126/science.aav1741).
#'
#' @return a `KdModel` object
#' @examples
#' data(SampleKdModel)
#' SampleKdModel
NULL

#' miRNA affinity models
#'
#' Methods for the \code{\link{KdModel}} class
#'
#' @name KdModel
#' @rdname KdModel
#' @aliases KdModel-methods KdModel-class
#' @seealso \code{\link{KdModel}}, \code{\link{KdModelList}}
#' @param object,x,... An object of class \code{\link{KdModel}}
#' @return Depends on the method.
#' @exportClass KdModel
#' @examples
#' data(SampleKdModel)
#' SampleKdModel
#' summary(SampleKdModel)
setClass(
  "KdModel",
  contains="list",
  validity=function(object){
    for(f in c("name","canonical.seed","mirseq")){
      if(!is.character(object[[f]]) || length(object[[f]])!=1)
        stop("The model should have a `",f,
             "` slot (character vector of length 1).")
    }
    if(is.null(object$mer8) || !(length(object$mer8) %in% c(1024,1440)) ||
       !is.integer(object$mer8)){
      stop("The `mer8` slot should be an integer vector of length 1024.")
    }
    if(is.null(object$fl) || !(length(object$fl) %in% c(1024,1440)) ||
       !is.integer(object$fl)){
      stop("The `fl` slot should be an integer vector of length 1024.")
    }
  }
)


#' @rdname KdModel
#' @export
setMethod("show", "KdModel", function(object){
  con <- conservation(object)
  con <- ifelse(is.na(con),"",paste0(" (",as.character(con),")"))
  cat(paste0("A `KdModel` for ", object$name, con, "\n  Sequence: ",
             gsub("T","U",object$mirseq), "\n  Canonical target seed: ",
             gsub("A$","(A)",object$canonical.seed)))
})


#' @rdname KdModel
#' @export
setMethod("summary", "KdModel", function(object){
  c( name=object$name, sequence=gsub("T","U",object$mirseq),
     canonical.seed=object$canonical.seed,
     conservation=as.character(conservation(object)) )
})

#' @rdname KdModel
#' @export
setMethod("c", signature(x = "KdModel"), function (x, ...){
  KdModelList(c(list(x), list(...)))
})


#' getKdModel
#'
#' @param kd A data.frame containing the log_kd per 12-mer sequence, or the
#' path to a text/csv file containing such a table. Should contain the columns
#' 'log_kd', '12mer' (or 'X12mer'), and eventually 'mirseq' (if the `mirseq`
#' argument is NULL) and 'mir' (if the `name` argument is NULL).
#' @param mirseq The miRNA (cDNA) sequence.
#' @param name The name of the miRNA.
#' @param conservation The conservation level of the miRNA. See
#' `scanMiR:::.conservation_levels()` for possible values.
#' @param ... Any additional information to be saved with the model.
#'
#' @return An object of class `KdModel`.
#' @export
#' @importFrom stats .lm.fit median cor
#' @importFrom utils read.delim
#' @importFrom methods is new
#' @examples
#' kd <- dummyKdData()
#' mod <- getKdModel(kd=kd, mirseq="TTAATGCTAATCGTGATAGGGGTT", name="my-miRNA")
getKdModel <- function( kd, mirseq=NULL, name=NULL, conservation=NA_integer_,
                        ...){
  if(is.character(kd) && length(kd)==1){
    if(is.null(name)) name <- gsub("\\.txt$|\\.csv$","",
                                   gsub("_kds","",basename(kd)))
    kd <- read.delim(kd, header=TRUE, stringsAsFactors=FALSE)[,c(1,2,4)]
  }
  if(is.null(mirseq) && !is.null(kd$mirseq))
    mirseq <- as.character(kd$mirseq[1])
  if(is.null(name) && !is.null(kd$mir)) name <- as.character(kd$mir[1])
  stopifnot(!is.null(name) && !is.null(mirseq))
  if(!("X12mer" %in% colnames(kd)) && "12mer" %in% colnames(kd))
    colnames(kd) <- gsub("^12mer$","X12mer",colnames(kd))
  kd <- kd[,c("X12mer","log_kd")]
  seed <- reverseComplement(DNAString(substr(mirseq, 2,8)))
  seed <- paste0(as.character(seed),"A")
  w <- grep("X|N",kd$X12mer,invert=TRUE)
  pwm <- Biostrings::consensusMatrix(
    as.character(rep(kd$X12mer[w], floor( (exp(-kd$log_kd[w]))/3 ))),
    as.prob=TRUE, width=12L
  )
  fields <- c("mer8","fl.score")
  if(!all(fields %in% colnames(kd))) kd <- .prep12mers(kd, seed=seed)
  fields <- c(fields, "log_kd")
  if(!all(fields %in% colnames(kd))) stop("Malformed `kd` data.frame.")
  co <- t(vapply(split(kd[,c("log_kd","fl.score")], kd$mer8),
                 FUN.VALUE=numeric(2), FUN=function(x){
    .lm.fit(cbind(1,x$fl.score),x$log_kd)$coefficients
  }))
  fitted <- co[kd$mer8,1]+co[kd$mer8,2]*kd$fl.score
  if(!is.na(conservation)){
    co <- .conservation_levels()
    if(!is.numeric(conservation)){
      if(!(conservation %in% co)){
        warning("Unknown conservation level - will be set to NA")
        conservation <- NA_integer_
      }else{
        conservation <- as.integer(names(co)[which(co==conservation)])
      }
    }else{
      if(!(as.character(conservation) %in% names(co)))
        warning("Unknown conservation level.")
    }
  }
  new("KdModel", list(mer8=as.integer(round(co[,1]*1000)),
                      fl=as.integer(round(co[,2]*1000)),
                      name=name, mirseq=mirseq, canonical.seed=seed,
                      pwm=pwm, conservation=conservation,
                      cor=cor(fitted, kd$log_kd),
                      mae=median(abs(kd$log_kd-fitted)), ... ))
}

.prep12mers <- function(x, seed){
  if(is.data.frame(x)){
    if(!all(c("X12mer","log_kd") %in% colnames(x)))
      stop("`x` should be a character vector or a data.frame with the columns ",
           "'X12mer' and 'log_kd'")
    x <- x[grep("N|X",substr(x$X12mer, 3,10),invert=TRUE),]
    x <- cbind(x[,"log_kd",drop=FALSE], .prep12mers(x$X12mer, seed=seed))
    return(x[!is.na(x$mer8),])
  }
  x <- gsub("X","N",as.character(x))
  y <- .getFlankingScore(x)
  data.frame(mer8=as.integer(factor(substr(x, 3,10),
                                    levels=getSeed8mers(seed))),
             fl.score=y$score, fl.ratio=y$ratio)
}

.flankingValues <- function(){
  matrix(c(-0.24, -0.14, 0, 0.1, 0.28, -0.24, -0.3, 0, 0.13, 0.42, -0.075,
           -0.18, 0, 0, 0.25, -0.1, -0.1, 0, 0, 0.26),
         nrow=5, dimnames=list(c("A","T","N","C","G")))
}

.getFlankingScore <- function(x){
  fl.s <- .flankingValues()
  fl.m <- cbind(substr(x,1,1), substr(x,2,2), substr(x,11,11),substr(x,12,12))
  fl.m <- matrix(as.integer(factor(fl.m, row.names(fl.s))), ncol=4)
  fl.score <- vapply(1:4, FUN.VALUE=numeric(length(x)),
                     FUN=function(i) fl.s[fl.m[,i,drop=FALSE],i,drop=FALSE])
  if(is.null(dim(fl.score))) fl.score <- matrix(fl.score, ncol=4)
  fl.score <- rowSums(fl.score)
  fl.ratio <- rowSums( (fl.m-3)>0 ) - rowSums( (fl.m-3)<0 )
  return( list(score=fl.score, ratio=fl.ratio) )
}

.add8merN <- function(mod, mer8=NULL){
  if(is.null(mer8)) mer8 <- getSeed8mers(mod$canonical.seed, addNs=TRUE)
  i1 <- split(1:1024, substr(mer8[1:1024],2,8))
  i2 <- split(1:1024, substr(mer8[1:1024],1,7))
  fl <- co <- rep(0L,416)
  m1 <- grep("^N",mer8)
  m2 <- grep("N$",mer8)
  tmpf <- function(x,i) as.integer(vapply(x, FUN.VALUE=numeric(1),
                                          FUN=function(j) median(mod[[i]][j])))
  co[m1-1024] <- tmpf(i1[gsub("N","",mer8[m1])], "mer8")
  fl[m1-1024] <- tmpf(i1[gsub("N","",mer8[m1])], "fl")
  co[m2-1024] <- tmpf(i2[gsub("N","",mer8[m2])], "mer8")
  fl[m2-1024] <- tmpf(i2[gsub("N","",mer8[m2])], "fl")
  mod$mer8 <- c(mod$mer8, co)
  mod$fl <- c(mod$fl, fl)
  mod
}

#' assignKdType
#'
#' Assigns a log_kd and match type to a set of matched sequences.
#'
#' @param x A vector of matched sequences, each of 12 nucleotides
#' @param mod An object of class `KdModel`
#' @param mer8 The optional set of 8mers included in the model (for internal
#' use; can be reconstructed from the model).
#'
#' @return A data.frame with one row for each element of `x`, and the columns
#' `type` and `log_kd`. To save space, the reported log_kd is multiplied by
#' 1000, rounded and saved as an integer.
#' @export
#' @examples
#' data(SampleKdModel)
#' assignKdType(c("CTAGCATTAAGT","ACGTACGTACGT"), SampleKdModel)
assignKdType <- function(x, mod, mer8=NULL){
  stopifnot(is(mod,"KdModel"))
  if(is.null(mer8)) mer8 <- getSeed8mers(mod$canonical.seed, addNs=TRUE)
  mod <- .add8merN(mod, mer8)
  fl.score <- as.numeric(.getFlankingScore(x)$score)
  mer9 <- factor(as.character(subseq(x, 2, 10)))
  mer8 <- factor(as.character(subseq(x, 3,10)), levels=mer8)
  d <- data.frame(
    type=getMatchTypes(levels(mer9), mod$canonical.seed)[as.integer(mer9)],
    log_kd=as.integer(round(mod$mer8[mer8] + fl.score*mod$fl[mer8]))
  )
  d$log_kd[is.na(d$log_kd)] <- 0L
  d
}

#' aggregateMatches
#'
#' Aggregates miRNA binding sites with log_kd values to predict transcript
#' repression. See the vignette for more detail.
#'
#' @param m A GRanges or data.frame of matches as returned by `findSeedMatches`.
#' @param a The relative concentration of unbound AGO-miRNA complexes.
#' @param b Factor specifying the additional repression by a single bound AGO.
#' @param c Penalty for sites that are found within the ORF region.
#' @param p3 Factor specifying additional repression due to 3p alignment.
#' @param coef_utr Factor specifying additional repression due to UTR length.
#' @param coef_orf Factor specifying additional repression due to ORF length.
#' @param p3.range Range used for 3p alignment.
#' @param keepSiteInfo Logical; whether to return information about site types
#' (default = TRUE). Ignored if `m` does not contain `log_kd` values
#' @param toInt Logical; whether to convert repression scores to integers
#' (default = FALSE).
#' @param BP Pass `BiocParallel::MulticoreParam(ncores, progressbar=TRUE)` to
#' enable multithreading. Note that in addition, `aggregateMatches` uses the
#' \link{data.table} package, which is often set to use multi-threading by
#' default (which would be multiplied by threads determined by `BP`). See
#' \code{\link[data.table]{setDTthreads}} for more information.
#'
#' @return a data.frame containing aggregated repression values and/or
#' information about the numbers and types of matches
#' @export
#' @importFrom stats quantile
#' @importFrom data.table as.data.table .N := dcast rbindlist .SD
#'
#' @examples
#' # we create mock RNA sequences and seeds:
#' seqs <- getRandomSeq(n=10)
#'
#' # load sample KdModel
#' data(SampleKdModel)
#'
#' # find matches
#' matches <- findSeedMatches(seqs, SampleKdModel)
#'
#' # aggregate matches
#' aggregateMatches(matches)
aggregateMatches <- function(m, a=0.007726 , b=0.5735, c=0.1810, p3=0.051,
                           coef_utr=0, coef_orf=0, p3.range=c(3L,8L),
                           keepSiteInfo = TRUE, toInt=FALSE, BP=NULL){
  if(is.null(BP)) BP <- BiocParallel::SerialParam()
  ll <- NULL # length info
  if(is(m,"GRanges")){
    if(length(m)==0) stop("Your search did not result in any match.")
    if(!is.null(ll <- metadata(m)$tx_info)){
      ll <- ll[,c("ORF.length", "UTR.length")]
      colnames(ll) <- tolower(colnames(ll))
      ll$transcript <- row.names(ll)
    }
    m$transcript <- as.factor(seqnames(m))
    m <- mcols(m)
    if(!is.null(m$miRNA)) m$miRNA <- droplevels(as.factor(m$miRNA))
    m <- as.data.frame(m)
  }else{
    if(nrow(m)==0) stop("Your search did not result in any match.")
    if(!is.null(ll <- attr(m, "tx_info"))){
      ll <- ll[,c("ORF.length", "UTR.length")]
      colnames(ll) <- tolower(colnames(ll))
      ll$transcript <- row.names(ll)
    }
  }
  if(!is.null(m$miRNA)){
    m <- split(m, m$miRNA)
    m <- bplapply(m, BPPARAM=BP, FUN=function(x){
      .aggregate_miRNA(x, ll=ll, a=a, b=b, c=c, p3=p3,coef_utr = coef_utr,
                       coef_orf = coef_orf, keepSiteInfo = keepSiteInfo,
                       toInt=toInt, p3.range=p3.range)
    })
    m <- as.data.frame(data.table::rbindlist(m, use.names=TRUE, fill=TRUE,
                                             idcol = "miRNA"))
    m$miRNA <- as.factor(m$miRNA)
    for(f in colnames(m)){
      if(is.numeric(m[[f]])) m[[f]][is.na(m[[f]])] <- 0L
    }
  }else{
    m <- .aggregate_miRNA(m, ll=ll, a=a, b=b, c=c, p3=p3,coef_utr = coef_utr,
                          coef_orf = coef_orf, keepSiteInfo = keepSiteInfo,
                          toInt=toInt, p3.range=p3.range)
  }
  return(m)
}


.aggregate_miRNA <- function(m, ll = NULL, a=0.007726, b=0.5735, c=0.1810,
                             p3=0.051, coef_utr = -0.17106,
                             coef_orf = -0.21546, p3.range=c(3L,8L),
                             keepSiteInfo = FALSE, toInt=FALSE){
  if(is(m,"GRanges")){
    m$transcript <- as.factor(seqnames(m))
    m <- mcols(m)
    if(!is.null(m$miRNA)) m$miRNA <- droplevels(as.factor(m$miRNA))
  }
  m <- as.data.table(m)
  if(is.null(m$log_kd)){
    m <- .aggregateSiteInfo(as.data.table(m))
    return(as.data.frame(m, stringsAsFactor=TRUE))
  }
  if(keepSiteInfo){
    m_type_table <- .aggregateSiteInfo(m)
  }
  for(col in c("ORF", "p3.score", "type")) if(is.null(m[[col]])) m[[col]] <- 0L
  if(!is.null(m$miRNA)){
    m <- m[,c("miRNA","transcript","ORF","log_kd","p3.score","type")]
  }else{
    m <- m[,c("transcript","ORF","log_kd","p3.score","type")]
  }
  m <- as.data.table(m)
  m[, ORF:=as.integer(ORF)]
  m[, log_kd:=-log_kd/1000]
  m <- m[log_kd>0]
  m$p3.score <- ifelse(m$type == "non-canonical" , 0L, m$p3.score)
  m$p3.score[m$p3.score>max(p3.range)] <- as.integer(max(p3.range))
  m$p3.score[m$p3.score<min(p3.range)] <- 0L
  m$N <- 1 / (1 + exp(-1 * (log(a) + m$log_kd + log(c)*m$ORF + p3*m$p3.score) ))
  m$log_kd <- NULL
  m$N_bg <- 1 / (1 + exp(-1 * (log(a)  + log(c)*m$ORF) ))
  m <- as.data.frame(rowsum(as.matrix(m[,c("N","N_bg")]), group=m$transcript))
  m <- data.frame( transcript=as.factor(row.names(m)),
                   repression=log2(1+exp(b)*m$N_bg) - log2(1 + exp(b)*m$N))
  if(!is.null(ll) && nrow(m) > 1){
    m <- merge(m,ll,by = "transcript", all.x = TRUE)

    # get the utr score
    m$utr.length <- log10(m$utr.length)
    m$utr.length[is.infinite(m$utr.length) | is.na(m$utr.length)] <- 0
    qu_un <- m[!duplicated(m$transcript),"utr.length"]
    qu <- quantile(qu_un, probs = c(0.05,0.95), na.rm = TRUE)
    m$utr_score <- (m$utr.length - qu[1]) / (qu[2] - qu[1])
    m$utr_score[is.na(m$utr_score)] <- 0

    # get the orf score
    if(sum(m$orf.length, na.rm = TRUE) > 0){
      m$orf.length <- log10(m$orf.length)
      m$orf.length[is.infinite(m$orf.length) | is.na(m$orf.length)] <- 0
      qu_un <- m[!duplicated(m$transcript),"orf.length"]
      qu <- quantile(qu_un, probs = c(0.05,0.95), na.rm = TRUE)
      m$orf_score <- (m$orf.length - qu[1]) / (qu[2] - qu[1])
      m$orf_score[is.na(m$orf_score)] <- 0
    }else{
      m$orf_score <- 0
    }
  m$repression <- m$repression + coef_utr*m$utr_score*m$repression +
    coef_orf*m$orf_score*m$repression
  m <- subset(m, select = - c(orf.length,utr.length,utr_score,orf_score))
  }
  if(toInt) m$repression <- as.integer(round(1000*m$repression))
  m$repression <- ifelse(m$repression >= 0, 0, m$repression)
  if(keepSiteInfo){
    if(!is.null(m$miRNA)){
      m <- merge(m, m_type_table, by=c("transcript","miRNA"), all=TRUE)
    }else{
      m <- merge(m, m_type_table, by=c("transcript"), all=TRUE)
    }
    m$miRNA <- NULL
  }
  m
}

.aggregateSiteInfo <- function(m) {
  hasORF <- !is.null(m$ORF)
  if(is.null(m$ORF)) m[,ORF:=0]
  if(is.null(m$miRNA)){
    if(any(m$ORF==0)) {
      sites <- dcast(m[ORF==0,.(N=.N), by=c("transcript","type")],
                      formula=transcript~type, value.var="N", fill=0L )
    } else {
      sites <- dcast(m[,.(N=.N), by=c("transcript","type")],
                     formula=transcript~type, value.var="N", fill=0L )
      cols <- names(sites)[vapply(sites, is.numeric, logical(1))]
      sites[, (cols):=0L]
    }
    if(hasORF) {
      if(any(m$ORF)==1) {
        sites_ORF <- dcast( m[ORF==1,.(N=.N), by=c("transcript","type")],
                        formula=transcript~type, value.var="N", fill=0L )
      } else {
        sites_ORF <- sites[, "transcript"]
      }
    }
  } else{
    if(any(m$ORF==0)) {
      sites <- dcast( m[ORF==0,.(N=.N), by=c("transcript","miRNA","type")],
                    formula=transcript+miRNA~type, value.var="N",
                    fill=0L )
    } else {
      sites <- dcast(m[,.(N=.N), by=c("transcript", "miRNA", "type")],
                     formula=transcript+miRNA~type, value.var="N", fill=0L )
      cols <- names(sites)[vapply(sites, is.numeric, logical(1))]
      sites[, (cols):=0L]
    }
    if(hasORF) {
      if(any(m$ORF)==1) {
        sites_ORF <- dcast( m[ORF==1,.(N=.N), by=c("transcript","miRNA","type")],
                      formula=transcript+miRNA~type, value.var="N",
                      fill=0L )
      } else {
        sites_ORF <- sites[, c("transcript", "miRNA")]
      }
    }
  }
  ind_bulged <- grep("bulged", names(sites))
  if(length(ind_bulged)>0) {
    ind_nc <- c(ind_bulged, grep("non-canonical", names(sites)))
    sites[,"non-canonical":= rowSums(.SD, na.rm = TRUE), .SDcols = ind_nc]
    sites[, grep("bulged", names(sites)):=NULL]
  }
  ind_wobbled <- grep("wobbled", names(sites))
  if(length(ind_wobbled)>0) {
    ind_nc2 <- c(ind_wobbled, grep("non-canonical", names(sites)))
    sites[,"non-canonical":= rowSums(.SD, na.rm = TRUE), .SDcols = ind_nc2]
    sites[, grep("wobbled", names(sites)):=NULL]
  }
  if(hasORF) {
    ind_bulged_orf <- grep("bulged", names(sites_ORF))
    if(length(ind_bulged_orf)>0) {
      ind_nc_orf <- c(ind_bulged_orf, grep("non-canonical", names(sites_ORF)))
      sites_ORF[,"non-canonical":= rowSums(.SD, na.rm = TRUE), .SDcols = ind_nc_orf]
      sites_ORF[, grep("bulged", names(sites_ORF)):=NULL]
    }
    ind_wobbled_orf <- grep("wobbled", names(sites_ORF))
    if(length(ind_wobbled_orf)>0) {
      ind_nc2_orf <- c(ind_wobbled_orf, grep("non-canonical", names(sites_ORF)))
      sites_ORF[,"non-canonical":= rowSums(.SD, na.rm = TRUE), .SDcols = ind_nc2_orf]
      sites_ORF[, grep("wobbled", names(sites_ORF)):=NULL]
    }
  }
  ind_6mer <- grep("6mer", names(sites))
  if(length(ind_6mer)>0) {
    sites[, "6mer" := rowSums(.SD, na.rm=TRUE), .SDcols = ind_6mer]
  }
  ind_7mer <- grep("7mer", names(sites))
  if(length(ind_7mer)>0) {
    sites[, "7mer" := rowSums(.SD, na.rm=TRUE), .SDcols = ind_7mer]
  }
  for(col in setdiff(c("8mer", "7mer", "6mer", "non-canonical"), names(sites))){
    sites[, (col):=0L]
  }
  if(hasORF) {
    cols <- names(sites_ORF)[vapply(sites_ORF, is.numeric, logical(1))]
    if(length(cols)==0) {
      sites_ORF[, ORF.canonical:=0L]
      sites_ORF[, ORF.nonCanonical:=0L]
    } else {
      sites_ORF[, "ORF.canonical":=rowSums(.SD, na.rm=TRUE), .SDcols = cols[which(cols != "non-canonical")]]
      sites_ORF[, "ORF.nonCanonical":=rowSums(.SD, na.rm=TRUE), .SDcols = cols[which(cols == "non-canonical")]]
      sites_ORF[, (cols):=NULL]
    }
  }
  if(is.null(m$miRNA)) {
    if(hasORF) {
      cols <- c("transcript", "8mer", "7mer", "6mer", "non-canonical",
                "ORF.canonical","ORF.nonCanonical")
      sites <- merge(sites, sites_ORF, by="transcript", all=TRUE)
    }
    else {
      cols <- c("transcript", "8mer", "7mer", "6mer", "non-canonical")
    }
  } else{
    if(hasORF) {
      cols <- c("transcript", "miRNA", "8mer", "7mer", "6mer", "non-canonical",
                "ORF.canonical","ORF.nonCanonical")
      sites <- merge(sites, sites_ORF, by=c("transcript", "miRNA"), all=TRUE)
    } else {
      cols <- c("transcript", "miRNA", "8mer", "7mer", "6mer", "non-canonical")
    }

  }
  for(i in names(sites)){
    sites[is.na(get(i)), (i):=0L]
    if(is.numeric(sites[[i]])) sites[[i]] <- as.integer(sites[[i]])
  }
  cols <- intersect(cols, colnames(sites))
  return(sites[,..cols])
}


.datatable.aware = TRUE

matches <- GRanges("seq",
                   IRanges(start = c(1,10), width = 7),
                   type = c("8mer", "7mer"),
                   log_kd = c(-5000, -3000),
                   p3.score = 8
)

agg <- aggregateMatches(matches, p3 = 0, coef_utr = 0, coef_orf = 0)

test_that("Aggregation format is correct", {
  expect_s3_class(agg, "data.frame")
  expect_identical(names(agg), c("transcript", "repression", "8mer", "7mer", 
                                 "6mer", "non-canonical"))
})

test_that("Predicted repression is correct", {
  expect_equal(agg$repression, -1.089705, tolerance = 1e-6)
})

test_that("Site info is correct", {
  expect_identical(agg$`8mer`, 1L)
  expect_identical(agg$`7mer`, 1L)
  expect_identical(agg$`6mer`, 0L)
  expect_identical(agg$`non-canonical`, 0L)
})

matches_df <- as.data.frame(matches)
names(matches_df)[1] <- "transcript"
agg_from_df <- aggregateMatches(matches_df, p3 = 0, coef_utr = 0, coef_orf = 0)

test_that("Aggregation also works for data.frame", {
  expect_identical(agg_from_df, agg)
})

matches_ORF <- matches
matches_ORF$ORF <- c(TRUE, FALSE)
agg_ORF <- aggregateMatches(matches_ORF, p3 = 0, coef_utr = 0, coef_orf = 0)
test_that("Predicted repression with ORF site is correct", {
  expect_equal(agg_ORF$repression, -0.6030502, tolerance = 1e-6)
  expect_identical(agg_ORF$`8mer`, 0L)
  expect_identical(agg_ORF$`7mer`, 1L)
  expect_identical(agg_ORF$`6mer`, 0L)
  expect_identical(agg_ORF$`non-canonical`, 0L)  
  expect_identical(agg_ORF$`ORF.canonical`, 1L)
})

agg_3p <- aggregateMatches(matches, p3 = 1, coef_utr = 0, coef_orf = 0)

test_that("Predicted repression with 3p score is correct", {
  expect_equal(agg_3p$repression, -2.14542, tolerance = 1e-6)
})

matches_noKD <- matches
matches_noKD$log_kd <- NULL
agg_noKD <- aggregateMatches(matches_noKD)

test_that("Site info is returned if no KDs are given", {
  expect_identical(agg_noKD, agg[,-2])
})

matches_miRNA <- c(matches_ORF, matches_ORF)
matches_miRNA$miRNA <- factor(rep(c("miRNA1", "miRNA2"), each = 2))
agg_miRNA <- aggregateMatches(matches_miRNA, p3 = 0, coef_utr = 0, coef_orf = 0)
test_that("Aggregation works with several miRNAs", {
  expect_identical(agg_miRNA$miRNA, factor(c("miRNA1", "miRNA2")))
  expect_identical(agg_miRNA[,-1], rbind(agg_ORF, agg_ORF))
})

matches_coef <- GRanges(seqnames = c("seq1", "seq2"), 
                        ranges(matches),
                        type = matches$type,
                        log_kd = matches$log_kd,
                        p3.score = matches$p3.score,
                        ORF = matches_ORF$ORF,
                        miRNA = "miRNA1"
                        )
metadata(matches_coef)$tx_info <- data.frame(ORF.length = c(10, 20),
                                              UTR.length = c(10, 20),
                                              row.names = c("seq1", "seq2"))
agg_coef <- aggregateMatches(matches_coef, p3 = 0, coef_utr = 1, coef_orf = 1)
test_that("Predicted repression with coef_utr and coef_orf is correct", {
  expect_equal(agg_coef$repression, c(-0.338184, -0.898957), tolerance = 1e-6)
})

---
title: "miRNA affinity models and the KdModel class"
author: 
- name: Pierre-Luc Germain
  affiliation:
    - D-HEST Institute for Neuroscience, ETH
    - Lab of Statistical Bioinformatics, UZH
- name: Michael Soutschek
  affiliation: Lab of Systems Neuroscience, D-HEST Institute for Neuroscience, ETH
- name: Fridolin Gross
  affiliation: Lab of Systems Neuroscience, D-HEST Institute for Neuroscience, ETH
package: scanMiR
output:
  BiocStyle::html_document
abstract: |
  This vignettes introduces the KdModel and KdModelList classes used for storing
  miRNA 12-mer affinities and predicting the dissociation constant of specific
  sites.
vignette: |
  %\VignetteIndexEntry{2_Kdmodels}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include=FALSE}
library(BiocStyle)
knitr::opts_chunk$set(crop = NULL)
```

# 12-mer dissociation rates

[McGeary, Lin et al. (2019)](https://dx.doi.org/10.1126/science.aav1741) used
RNA bind-n-seq (RBNS) to empirically determine the affinities (i.e. dissoiation 
rates) of selected miRNAs towards random 12-nucleotide sequences (termed 
12-mers). As expected, bound sequences typically exhibited complementarity to 
the miRNA seed region (positions 2-8 from the miRNA's 5' end), but the study 
also revealed non-canonical bindings and the importance of flanking 
di-nucleotides. Based on these data, the authors developed a model which 
predicted 12-mer dissociation rates (KD) based on the miRNA sequence. ScanMiR
encodes a compressed version of these prediction in the form of a `KdModel`
object.

The 12-mer is defined as the 8 nucleotides opposite the miRNA's extended seed 
region plus flanking dinucleotides on either side:

```{r echo=FALSE, out.width="35%", fig.align = 'center'}
knitr::include_graphics(system.file('docs', '12mer.png', package = 'scanMiR'))
```


# KdModels

The `KdModel` class contains the information concerning the sequence (12-mer) 
affinity of a given miRNA, and is meant to compress and make easily manipulable 
the dissociation constants (Kd) predictions from 
[McGeary, Lin et al. (2019)](https://dx.doi.org/10.1126/science.aav1741).

We can take a look at the example `KdModel`:

```{r}
library(scanMiR)
data(SampleKdModel)
SampleKdModel
```

In addition to the information necessary to predict the binding affinity to any 
given 12-mer sequence, the model contains, minimally, the name and sequence of 
the miRNA. Since the `KdModel` class extends the list class, any further 
information can be stored:

```{r}
SampleKdModel$myVariable <- "test"
```

An overview of the binding affinities can be obtained with the following plot:

```{r}
plotKdModel(SampleKdModel, what="seeds")
```

The plot gives the -log(Kd) values of the top 7-mers (including both canonical 
and non-canonical sites), with or without the final "A" vis-à-vis the first 
miRNA nucleotide.

To predict the dissociation constant (and binding type, if any) of a given 
12-mer sequence, you can use the `assignKdType` function:

```{r}
assignKdType("ACGTACGTACGT", SampleKdModel)
# or using multiple sequences:
assignKdType(c("CTAGCATTAAGT","ACGTACGTACGT"), SampleKdModel)
```

The log_kd column contains log(Kd) values multiplied by 1000 and stored as an 
integer (which is more economical when dealing with millions of sites). In the 
example above, `r (lkd <- assignKdType("CTAGCATTAAGT", SampleKdModel)$log_kd)` 
means `r lkd/1000`, or a dissociation constant of `r exp(lkd/1000)`. The 
smaller the values, the stronger the relative affinity.

## KdModelLists

A `KdModelList` object is simply a collection of `KdModel` objects. We can 
build one in the following way:

```{r}
# we create a copy of the KdModel, and give it a different name:
mod2 <- SampleKdModel
mod2$name <- "dummy-miRNA"
kml <- KdModelList(SampleKdModel, mod2)
kml
summary(kml)
```

Beyond operations typically performed on a list (e.g. subsetting), some 
specific slots of the respective KdModels can be accessed, for example: 

```{r}
conservation(kml)
```

# Creating a KdModel object

`KdModel` objects are meant to be created from a table assigning a log_kd 
values to 12-mer target sequences, as produced by the CNN from McGeary, Lin et 
al. (2019). For the purpose of example, we create such a dummy table:

```{r}
kd <- dummyKdData()
head(kd)
```

A `KdModel` object can then be created with:

```{r}
mod3 <- getKdModel(kd=kd, mirseq="TTAATGCTAATCGTGATAGGGGTT", name = "my-miRNA")
```

Alternatively, the `kd` argument can also be the path to the output file of the 
CNN (and if `mirseq` and `name` are in the table, they can be omitted).

# Common KdModel collections

The [scanMiRData](https://github.com/ETHZ-INS/scanMiRData) package contains 
`KdModel` collections corresponding to all human, mouse and rat mirbase miRNAs.

# Under the hood

When calling `getKdModel`, the dissociation constants are stored as an 
lightweight overfitted linear model, with base KDs coefficients (stored as 
integers in `object$mer8`) for each 1024 partially-matching 8-mers (i.e. at 
least 4 consecutive matching nucleotides) to which are added 8-mer-specific 
coefficients (stored in `object$fl`) that are multiplied with a flanking score 
generated by the flanking di-nucleotides. The flanking score is calculated 
based on the di-nucleotide effects experimentally measured by McGeary, Lin et 
al. (2019). To save space, the actual 8-mer sequences are not stored but 
generated when needed in a deterministic fashion. The 8-mers can be obtained, 
in the right order, with the `getSeed8mers` function.

<br/><br/>

# Session info {.unnumbered}

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plotKdModel}
\alias{plotKdModel}
\title{plotKdModel}
\usage{
plotKdModel(mod, what = c("both", "seeds", "logo"), n = 10)
}
\arguments{
\item{mod}{A `KdModel`}

\item{what}{Either 'seeds', 'logo', or 'both' (default).}

\item{n}{The number of top 7-mers to plot (when `what='seeds'`)}
}
\value{
If `what="logo"`, returns nothing and plots a position weight matrix.
Otherwise returns a ggplot.
}
\description{
Plots the summary of an affinity model.
}
\details{
`what='seeds'` plots the -$log(K_d)$ values of the top `n` 7-mers (including
both canonical and non-canonical sites), with or without the final "A"
vis-a-vis the first miRNA nucleotide.
`what='logo'` plots a `seqLogo` (requires the
[seqLogo]{https://bioconductor.org/packages/release/bioc/html/seqLogo.html}
package) showing the nucleotide-wise information content and preferences for
all 12-mers (centered around the seed, oriented in the direction of the target
mRNA). `what="both"` plots both.
Note that if the package `ggseqlogo` is installed, this will be used instead
to plot the logo, resulting in more detailed plot annotation.
}
\examples{
data(SampleKdModel)
plotKdModel(SampleKdModel, what="seeds")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KdModelList.R
\name{KdModelList-methods}
\alias{KdModelList-methods}
\alias{KdModelList-methods,KdModelList-method}
\alias{summary,KdModelList-method}
\alias{[,KdModelList,ANY-method}
\alias{[,KdModelList-methods,KdModelList-method}
\title{Methods for the \code{\link{KdModelList}} classes}
\usage{
\S4method{summary}{KdModelList}(object)

\S4method{[}{KdModelList,ANY}(x, i, j = NULL, ..., drop = TRUE)
}
\arguments{
\item{object, x}{An object of class \code{\link{KdModelList}}}

\item{i}{the index of item(s) to select}

\item{j, drop, ...}{ignored}
}
\value{
Depends on the method.
}
\description{
Methods for the \code{\link{KdModelList}} classes
}
\examples{
# create a KdModelList :
data(SampleKdModel)
kml <- KdModelList( SampleKdModel, SampleKdModel, makeUnique=TRUE )
summary(kml)
kml[1] # returns a KdModelList
kml[[2]] # returns a KdModel
conservation(kml)
}
\seealso{
\code{\link{KdModel}}, \code{\link{KdModelList}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{getRandomSeq}
\alias{getRandomSeq}
\title{getRandomSeq}
\usage{
getRandomSeq(length = 3000, alphabet = c("A", "C", "G", "T"), n = 1)
}
\arguments{
\item{length}{Length of the sequence}

\item{alphabet}{Letters from which to sample}

\item{n}{The number of sequences to generate}
}
\value{
A character vector of length 1
}
\description{
Produces a random sequence of the given letters
}
\examples{
getRandomSeq(100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanning.R
\name{findSeedMatches}
\alias{findSeedMatches}
\title{Predicting and characterizing miRNA binding sites}
\usage{
findSeedMatches(
  seqs,
  seeds,
  shadow = 0L,
  onlyCanonical = FALSE,
  maxLogKd = c(-1, -1.5),
  keepMatchSeq = FALSE,
  minDist = 7L,
  p3.extra = FALSE,
  p3.params = list(maxMirLoop = 7L, maxTargetLoop = 9L, maxLoopDiff = 4L, mismatch =
    TRUE, GUwob = TRUE),
  agg.params = .defaultAggParams(),
  ret = c("GRanges", "data.frame", "aggregated"),
  BP = NULL,
  verbose = NULL,
  n_seeds = NULL,
  useTmpFiles = FALSE,
  keepTmpFiles = FALSE
)
}
\arguments{
\item{seqs}{A character vector or `DNAStringSet` of DNA sequences in which to
look.}

\item{seeds}{A character vector of 7-nt seeds to look for. If RNA, will be
reversed and complemented before matching. If DNA, they are assumed to be
the target sequence to look for. Alternatively, a list of objects of class
`KdModel` or an object of class `KdModelList` can be given.}

\item{shadow}{Integer giving the shadow, i.e. the number of nucleotides
hidden at the beginning of the sequence (default 0).}

\item{onlyCanonical}{Logical; whether to restrict the search only to
canonical binding sites.}

\item{maxLogKd}{Maximum log_kd value to keep. This has a major impact on the
number of sites returned, and hence on the memory requirements. Set to Inf
to disable (_not_ recommended when running large scans!).}

\item{keepMatchSeq}{Logical; whether to keep the sequence (including flanking
dinucleotides) for each seed match (default FALSE).}

\item{minDist}{Integer specifying the minimum distance between matches of the
same miRNA (default 7). Closer matches will be reduced to the
highest-affinity. To disable the removal of overlapping features, use
`minDist=-Inf`.}

\item{p3.extra}{Logical; whether to keep extra information about 3'
alignment. Disable (default) this when running large scans, otherwise you
might hit your system's memory limits.}

\item{p3.params}{Named list of parameters for 3' alignment with slots
`maxMirLoop` (integer, default = 7), `maxTargetLoop` (integer, default = 9),
`maxLoopDiff` (integer, default = 4), `mismatch`
(logical, default = TRUE) and `GUwob` (logical, default = TRUE).}

\item{agg.params}{A named list with slots `a`, `b`, `c`, `p3`, `coef_utr`,
`coef_orf` and `keepSiteInfo` indicating the parameters for the aggregation.
Ignored if `ret!="aggregated"`. For further details see documentation of
`aggregateMatches`.}

\item{ret}{The type of data to return, either "GRanges" (default),
"data.frame", or "aggregated" (aggregates affinities/sites for each
seed-transcript pair).}

\item{BP}{Pass `BiocParallel::MulticoreParam(ncores, progressbar=TRUE)` to
enable multithreading.}

\item{verbose}{Logical; whether to print additional progress messages
(default on if not multithreading)}

\item{n_seeds}{Integer; the number of seeds that are processed in parallel to
avoid memory issues.}

\item{useTmpFiles}{Logical; whether to write results for single miRNAs in
temporary files (ignored when scanning for a single seed). Alternatively,
`useTmpFiles` can be a character vector of length 1 indicating the path to
the directory in which to write temporary files.}

\item{keepTmpFiles}{Logical; whether to keep the temporary files at the end
of the process; ignored if `useTmpFiles=FALSE`. Temporary files are removed
only upon successful completion of the function, meaning that they will not
be deleted in case of errors.}
}
\value{
A GRanges of all matches. If `seeds` is a `KdModel` or `KdModelList`,
the `log_kd` column will report the ln(Kd) multiplied by 1000, rounded and
saved as an integer. If `ret!="GRanges`, returns a data.frame.
}
\description{
`findSeedMatches` takes a set of sequences and a set of miRNAs (given either
as target seeds, mature miRNA sequences, or a \code{\link{KdModelList}}).
}
\examples{
# we create mock RNA sequences and seeds:
seqs <- getRandomSeq(n=10)
seeds <- c("AAACCAC", "AAACCUU")
findSeedMatches(seqs, seeds)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{SampleKdModel}
\alias{SampleKdModel}
\title{Example KdModel (hsa-miR-155-5p)}
\value{
a `KdModel` object
}
\description{
`KdModel` for hsa-miR-155-5p, based on Kd predictions from the CNN of
[McGeary, Lin et al. (2019)](https://dx.doi.org/10.1126/science.aav1741).
}
\examples{
data(SampleKdModel)
SampleKdModel
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanning.R
\name{getMatchTypes}
\alias{getMatchTypes}
\title{getMatchTypes}
\usage{
getMatchTypes(x, seed, checkWobble = TRUE)
}
\arguments{
\item{x}{A character vector of short sequences.}

\item{seed}{A 7 or 8 nucleotides string indicating the seed (5' to 3'
sequence of the target RNA). If of length 7, an "A" will be appended.}

\item{checkWobble}{Whether to flag wobbled sites}
}
\value{
A factor of match types.
}
\description{
Given a seed and a set of sequences matching it, returns the type of match.
}
\examples{
x <- c("AACACTCCAG","GACACTCCGC","GTACTCCAT","ACGTACGTAC")
getMatchTypes(x, seed="ACACTCCA")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{viewTargetAlignment}
\alias{viewTargetAlignment}
\title{viewTargetAlignment}
\usage{
viewTargetAlignment(
  m,
  miRNA,
  seqs = NULL,
  flagBulgeMatches = FALSE,
  p3.params = list(),
  min3pMatch = 3L,
  hideSingletons = FALSE,
  UGsub = TRUE,
  ...,
  outputType = c("print", "data.frame", "plot", "ggplot")
)
}
\arguments{
\item{m}{A GRanges of length 1 giving the information for a given match, as
produced by \code{\link{findSeedMatches}}.}

\item{miRNA}{A miRNA sequence, or a \code{\link{KdModel}} object of the miRNA
corresponding to the match in `m`; alternatively, a \code{\link{KdModelList}}
including the model.}

\item{seqs}{The sequences corresponding to the seqnames of `m`. Not needed if
`m` contains the target sequences.}

\item{flagBulgeMatches}{Logical; whether to flag matches inside the bulge
(default FALSE)}

\item{p3.params}{See \code{\link{findSeedMatches}}.}

\item{min3pMatch}{The minimum 3' alignment for any to be plotted}

\item{hideSingletons}{Logical; whether to hide isolated single base-pair
matches}

\item{UGsub}{Logical; whether to show U-G matches}

\item{...}{Passed to `text` if `outputType="plot"`.}

\item{outputType}{Either 'print' (default, prints to console), 'data.frame',
or 'plot'.}
}
\value{
Returns nothing `outputType="print"`. If `outputType="data.frame"`,
returns a data.frame containing the alignment strings; if
`outputType="ggplot"` returns a `ggplot` object.
}
\description{
viewTargetAlignment
}
\examples{
data(SampleKdModel)
seq <- c(seq1="CGACCCCTATCACGTCCGCAGCATTAAAT")
m <- findSeedMatches(seq, SampleKdModel, verbose=FALSE)
viewTargetAlignment(m, miRNA=SampleKdModel, seqs=seq)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanning.R
\name{removeOverlappingRanges}
\alias{removeOverlappingRanges}
\title{removeOverlappingRanges}
\usage{
removeOverlappingRanges(
  x,
  minDist = 7L,
  retIndices = FALSE,
  ignore.strand = FALSE
)
}
\arguments{
\item{x}{A GRanges, sorted by (decreasing) importance.}

\item{minDist}{Minimum distance between ranges.}

\item{retIndices}{Logical; whether to return the indices of entries to
remove, rather than the filtered GRanges.}

\item{ignore.strand}{Logical. Whether the strand of the input ranges should
be ignored or not.}
}
\value{
A filtered GRanges, or an integer vector of indices to be removed if
`retIndices==TRUE`.
}
\description{
Removes elements from a GRanges that overlap (or are within a given distance
of) other elements higher up in the list (i.e. assumes that the ranges are
sorted in order of priority). The function handles overlaps between more than
two ranges by successively removing those that overlap higher-priority ones.
}
\examples{
library(GenomicRanges)
gr <- GRanges(seqnames=rep("A",4), IRanges(start=c(10,25,45,35), width=6))
removeOverlappingRanges(gr, minDist=7)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{SampleTranscript}
\alias{SampleTranscript}
\title{Example transcript sequence}
\value{
a named character vector of length 1.
}
\description{
An artificial transcript sequence used for examples.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KdModels.R
\name{assignKdType}
\alias{assignKdType}
\title{assignKdType}
\usage{
assignKdType(x, mod, mer8 = NULL)
}
\arguments{
\item{x}{A vector of matched sequences, each of 12 nucleotides}

\item{mod}{An object of class `KdModel`}

\item{mer8}{The optional set of 8mers included in the model (for internal
use; can be reconstructed from the model).}
}
\value{
A data.frame with one row for each element of `x`, and the columns
`type` and `log_kd`. To save space, the reported log_kd is multiplied by
1000, rounded and saved as an integer.
}
\description{
Assigns a log_kd and match type to a set of matched sequences.
}
\examples{
data(SampleKdModel)
assignKdType(c("CTAGCATTAAGT","ACGTACGTACGT"), SampleKdModel)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KdModelList.R
\name{conservation}
\alias{conservation}
\title{conservation}
\usage{
conservation(x)
}
\arguments{
\item{x}{A KdModelList, or a KdModel}
}
\value{
A vector of the conservation status for each miRNA
}
\description{
conservation
}
\examples{
data(SampleKdModel)
conservation(SampleKdModel)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{get8merRange}
\alias{get8merRange}
\title{get8merRange}
\usage{
get8merRange(mod)
}
\arguments{
\item{mod}{A `KdModel`}
}
\value{
A numeric vector of length two
}
\description{
Returns the minimum and maximum 8-mer log-kd values
}
\examples{
data("SampleKdModel")
get8merRange(SampleKdModel)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanning.R
\name{get3pAlignment}
\alias{get3pAlignment}
\title{Finds 3' complementary binding of a miRNA}
\usage{
get3pAlignment(
  seqs,
  mirseq,
  mir3p.start = 9L,
  allow.mismatch = TRUE,
  maxMirLoop = 7L,
  maxTargetLoop = 9L,
  maxLoopDiff = 4L,
  TGsub = TRUE,
  siteType = NULL
)
}
\arguments{
\item{seqs}{A set of sequences in which to look for 3' matches (i.e. upstream
of the seed match)}

\item{mirseq}{The sequence of the mature miRNA}

\item{mir3p.start}{The position in `mirseq` in which to start looking}

\item{allow.mismatch}{Logical; whether to allow mismatches}

\item{maxMirLoop}{Maximum miRNA loop size}

\item{maxTargetLoop}{Maximum target loop size}

\item{maxLoopDiff}{Maximum size difference between miRNA and target loops}

\item{TGsub}{Logical; whether to allow T/G substitutions.}

\item{siteType}{The optional type of seed-complementarity, as returned by
\code{\link{getMatchTypes}}. This is needed to identify slicing/TDMD sites.
If given, should be a vector of the same length as `seqs`.}
}
\value{
A data.frame with one row for each element of `seqs`, indicating the
size of the miRNA bulge, the size of the target mRNA bulge, the number of
mismatches at the 3' end, and the partial 3' alignment score (i.e. roughly
the number of consecutive matching nucleotides)
}
\description{
Performs a local alignment of the miRNA 3' sequence (determined by
`mir3p.start`) on given the given sequences.
}
\examples{
get3pAlignment(seqs="NNAGTGTGCCATNN", mirseq="TGGAGTGTGACAATGGTGTTTG")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregation.R
\name{aggregateMatches}
\alias{aggregateMatches}
\title{aggregateMatches}
\usage{
aggregateMatches(
  m,
  a = 0.007726,
  b = 0.5735,
  c = 0.181,
  p3 = 0.051,
  coef_utr = 0,
  coef_orf = 0,
  p3.range = c(3L, 8L),
  keepSiteInfo = TRUE,
  toInt = FALSE,
  BP = NULL
)
}
\arguments{
\item{m}{A GRanges or data.frame of matches as returned by `findSeedMatches`.}

\item{a}{The relative concentration of unbound AGO-miRNA complexes.}

\item{b}{Factor specifying the additional repression by a single bound AGO.}

\item{c}{Penalty for sites that are found within the ORF region.}

\item{p3}{Factor specifying additional repression due to 3p alignment.}

\item{coef_utr}{Factor specifying additional repression due to UTR length.}

\item{coef_orf}{Factor specifying additional repression due to ORF length.}

\item{p3.range}{Range used for 3p alignment.}

\item{keepSiteInfo}{Logical; whether to return information about site types
(default = TRUE). Ignored if `m` does not contain `log_kd` values}

\item{toInt}{Logical; whether to convert repression scores to integers
(default = FALSE).}

\item{BP}{Pass `BiocParallel::MulticoreParam(ncores, progressbar=TRUE)` to
enable multithreading. Note that in addition, `aggregateMatches` uses the
\link{data.table} package, which is often set to use multi-threading by
default (which would be multiplied by threads determined by `BP`). See
\code{\link[data.table]{setDTthreads}} for more information.}
}
\value{
a data.frame containing aggregated repression values and/or
information about the numbers and types of matches
}
\description{
Aggregates miRNA binding sites with log_kd values to predict transcript
repression. See the vignette for more detail.
}
\examples{
# we create mock RNA sequences and seeds:
seqs <- getRandomSeq(n=10)

# load sample KdModel
data(SampleKdModel)

# find matches
matches <- findSeedMatches(seqs, SampleKdModel)

# aggregate matches
aggregateMatches(matches)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{dummyKdData}
\alias{dummyKdData}
\title{Create dummy log_kd per 12-mer data}
\usage{
dummyKdData(mod = NULL)
}
\arguments{
\item{mod}{Optional model from which to create the dummy data}
}
\value{
A data.frame with 12-mers and log_kds
}
\description{
Create dummy log_kd per 12-mer data
}
\examples{
kd <- dummyKdData()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KdModels.R
\name{getKdModel}
\alias{getKdModel}
\title{getKdModel}
\usage{
getKdModel(kd, mirseq = NULL, name = NULL, conservation = NA_integer_, ...)
}
\arguments{
\item{kd}{A data.frame containing the log_kd per 12-mer sequence, or the
path to a text/csv file containing such a table. Should contain the columns
'log_kd', '12mer' (or 'X12mer'), and eventually 'mirseq' (if the `mirseq`
argument is NULL) and 'mir' (if the `name` argument is NULL).}

\item{mirseq}{The miRNA (cDNA) sequence.}

\item{name}{The name of the miRNA.}

\item{conservation}{The conservation level of the miRNA. See
`scanMiR:::.conservation_levels()` for possible values.}

\item{...}{Any additional information to be saved with the model.}
}
\value{
An object of class `KdModel`.
}
\description{
getKdModel
}
\examples{
kd <- dummyKdData()
mod <- getKdModel(kd=kd, mirseq="TTAATGCTAATCGTGATAGGGGTT", name="my-miRNA")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KdModelList.R
\docType{class}
\name{KdModelList-class}
\alias{KdModelList-class}
\alias{KdModelList}
\title{KdModelList}
\usage{
KdModelList(..., description = NULL, makeUnique = FALSE)
}
\arguments{
\item{...}{Any number of \code{\link{KdModel}} objects or lists thereof.}

\item{description}{A description for the collection.}

\item{makeUnique}{Logical; whether to rename models if names are duplicated.}
}
\value{
A KdModelList
}
\description{
KdModelList
}
\examples{
data(SampleKdModel)
mods <- KdModelList(SampleKdModel, SampleKdModel, makeUnique = TRUE)
mods
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KdModels.R
\docType{class}
\name{KdModel}
\alias{KdModel}
\alias{KdModel-methods}
\alias{KdModel-class}
\alias{show,KdModel-method}
\alias{summary,KdModel-method}
\alias{c,KdModel-method}
\title{miRNA affinity models}
\usage{
\S4method{show}{KdModel}(object)

\S4method{summary}{KdModel}(object)

\S4method{c}{KdModel}(x, ...)
}
\arguments{
\item{object, x, ...}{An object of class \code{\link{KdModel}}}
}
\value{
Depends on the method.
}
\description{
Methods for the \code{\link{KdModel}} class
}
\examples{
data(SampleKdModel)
SampleKdModel
summary(SampleKdModel)
}
\seealso{
\code{\link{KdModel}}, \code{\link{KdModelList}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{getSeed8mers}
\alias{getSeed8mers}
\title{getSeed8mers}
\usage{
getSeed8mers(seed, addNs = FALSE)
}
\arguments{
\item{seed}{The miRNA seed (target DNA sequence), a character vector of
length 8 (if of length 7, a "A" will be added on the right)}

\item{addNs}{Logical; whether to include 8mers with one flanking N}
}
\value{
A vector of 1024 8mers.
}
\description{
Generates all possible 8mers with 4 consecutive and positioned matches to a
given seed.
}
\examples{
head(getSeed8mers("ACACTCCA"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{getKmers}
\alias{getKmers}
\title{getKmers}
\usage{
getKmers(n = 4, from = c("A", "C", "G", "T"))
}
\arguments{
\item{n}{Number of elements}

\item{from}{Letters sampled}
}
\value{
A character vector
}
\description{
Returns all combinations of `n` elements of `from`
}
\examples{
getKmers(3)
}
